from lalamo.model_import.decoder_configs import HFLFM2Config
from lalamo.quantization import QuantizationMode

from .common import ConfigMap, FileSpec, ModelSpec

__all__ = ["LFM2_MODELS"]


def _lfm2_repo(size: str, quantization: QuantizationMode | None) -> tuple[str, str]:
    organization = "LiquidAI" if quantization is None else "mlx-community"
    name = f"LFM2-{size}{f'-{quantization.bits}bit' if quantization is not None else ''}"
    return (organization, name)


LFM2_MODELS = [
    ModelSpec(
        vendor="LiquidAI",
        family="LFM2",
        name=_lfm2_repo(size, quantization)[1],
        size=size,
        repo="/".join(_lfm2_repo(size, quantization)),
        config_type=HFLFM2Config,
        quantization=quantization,
        configs=ConfigMap(
            chat_template=FileSpec("chat_template.jinja"),
        ),
        use_cases=tuple(),
    )
    for size in ["350M", "700M", "1.2B", "2.6B"]
    for quantization in [None, *([QuantizationMode.UINT4, QuantizationMode.UINT8] if size != "2.6B" else [])]
]



import pytest

from lalamo.model_import.model_specs.lfm2 import LFM2_MODELS
from tests.lfm2_tracer import LFM2DecoderTracer
from tests.test_models import DType, ModelTestSpec, _test_model

MODEL_LIST = [ModelTestSpec(model.repo, DType.FLOAT32) for model in LFM2_MODELS if model.quantization is None]


@pytest.mark.parametrize("test_spec", MODEL_LIST, ids=[m.model_repo for m in MODEL_LIST])
def test_lfm2_models(test_spec: ModelTestSpec) -> None:
    _test_model(test_spec, LFM2DecoderTracer)




from lalamo.model_import.decoder_configs import HFLlamaConfig

from .common import ModelSpec

__all__ = ["PLEIAS_MODELS"]


PLEIAS_MODELS = [
    ModelSpec(
        vendor="PleIAs",
        family="Pleias-RAG",
        name="Pleias-RAG-1B",
        size="1B",
        quantization=None,
        repo="PleIAs/Pleias-RAG-1B",
        config_type=HFLlamaConfig,
    ),
]



from collections.abc import Mapping
from dataclasses import dataclass, replace
from typing import Self

import equinox as eqx
from jax import vmap
from jaxtyping import Array, DTypeLike, Float, Int, PRNGKeyArray

from lalamo.common import ParameterTree
from lalamo.modules.common import PositionalEmbeddingSelector
from lalamo.modules.linear import LinearBase, LinearConfig
from lalamo.modules.rope import PositionalEmbeddings

from .common import TokenMixerBase, TokenMixerConfigBase, TokenMixerResult
from .mamba import SeparableCausalConv, SeparableCausalConvConfig
from .state import ShortConvStateLayer

__all__ = [
    "ShortConv",
    "ShortConvConfig",
    "ShortConvResult",
]


ShortConvResult = TokenMixerResult[ShortConvStateLayer]


@dataclass(frozen=True)
class ShortConvConfig(TokenMixerConfigBase):
    in_projection_config: LinearConfig
    conv_config: SeparableCausalConvConfig
    out_projection_config: LinearConfig

    kernel_size: int

    @property
    def rope_dim(self) -> None:
        return None

    def random_init(
        self,
        model_dim: int,
        *,
        key: PRNGKeyArray,
    ) -> "ShortConv":
        in_projection = self.in_projection_config.random_init(
            input_dim=model_dim,
            output_dims=(model_dim,)*3,
            has_biases=False,
            key=key,
        )

        conv = self.conv_config.random_init(model_dim, self.kernel_size, key=key)

        out_projection = self.out_projection_config.random_init(
            input_dim=model_dim,
            output_dims=(model_dim,),
            has_biases=False,
            key=key,
        )

        return ShortConv(
            self,
            in_projection=in_projection,
            conv=conv,
            out_projection=out_projection,
        )

    def empty(
        self,
        model_dim: int,
    ) -> "ShortConv":
        in_projection = self.in_projection_config.empty(
            input_dim=model_dim,
            output_dims=(model_dim,)*3,
            has_biases=False,
        )

        conv = self.conv_config.empty(model_dim, self.kernel_size)

        out_projection = self.out_projection_config.empty(
            input_dim=model_dim,
            output_dims=(model_dim,),
            has_biases=False,
        )

        return ShortConv(
            self,
            in_projection=in_projection,
            conv=conv,
            out_projection=out_projection,
        )


class ShortConv(TokenMixerBase[ShortConvConfig, ShortConvStateLayer]):
    in_projection: LinearBase
    conv: SeparableCausalConv
    out_projection: LinearBase

    @property
    def activation_precision(self) -> DTypeLike:
        return self.in_projection.activation_precision

    @property
    def model_dim(self) -> int:
        return self.in_projection.input_dim

    @property
    def positional_embedding_selector(self) -> PositionalEmbeddingSelector:
        return PositionalEmbeddingSelector.NONE

    @eqx.filter_jit
    def __call__(
        self,
        inputs: Float[Array, "suffix_tokens channels"],
        positional_embeddings: PositionalEmbeddings | None,
        state: ShortConvStateLayer | None = None,
        return_updated_state: bool = False,
        length_without_padding: Int[Array, ""] | int | None = None, # noqa: ARG002
    ) -> TokenMixerResult[ShortConvStateLayer]:
        if positional_embeddings is not None:
            raise ValueError("Positional embeddings are not supported for ShortConv.")

        pre_conv_gate, post_conv_gate, x = vmap(self.in_projection)(inputs)

        prev_conv_state = state.conv_state if state is not None else None
        conv_output = self.conv(x * pre_conv_gate, prev_conv_state, return_updated_state)

        (outputs,) = vmap(self.out_projection)(conv_output.outputs * post_conv_gate)
        updated_conv_state = conv_output.state

        if return_updated_state:
            assert updated_conv_state is not None
            updated_state = ShortConvStateLayer(updated_conv_state)
        else:
            updated_state = None

        return TokenMixerResult(outputs, updated_state)

    def init_static_state(self, capacity: int) -> ShortConvStateLayer: # noqa: ARG002
        return ShortConvStateLayer.init(
            self.config.kernel_size,
            self.in_projection.input_dim,
            self.activation_precision,
        )

    def export_weights(self) -> ParameterTree:
        return {
            "in_projection": self.in_projection.export_weights(),
            "conv": self.conv.export_weights(),
            "out_projection": self.out_projection.export_weights(),
        }

    def import_weights(
        self,
        weights: ParameterTree[Array],
    ) -> Self:
        assert isinstance(weights, Mapping)
        assert isinstance(weights["in_projection"], Mapping)
        assert isinstance(weights["conv"], Mapping)
        assert isinstance(weights["out_projection"], Mapping)

        return replace(
            self,
            in_projection=self.in_projection.import_weights(weights["in_projection"]),
            conv=self.conv.import_weights(weights["conv"]),
            out_projection=self.out_projection.import_weights(weights["out_projection"]),
        )



from collections.abc import Callable, Iterable

import equinox as eqx
from jax._src.api import ShapeDtypeStruct
from jax.tree import leaves_with_path
from jax.tree_util import keystr
from jaxtyping import Array, PyTree

__all__ = [
    "load_parameters",
]


def _get_name(leaf: PyTree, tree: PyTree) -> str:
    for path, value in leaves_with_path(tree):
        if value is leaf:
            return f"~{keystr(path)}"
    raise ValueError(f"Leaf {leaf} not found in tree {tree}")


def _check_compatible(old_value: PyTree, new_value: PyTree, module: eqx.Module) -> None:
    if isinstance(old_value, (Array, ShapeDtypeStruct)) and isinstance(new_value, Array):
        name = _get_name(old_value, module)
        if old_value.shape != new_value.shape:
            raise ValueError(f"Expected parameter {name} to have shape {old_value.shape}, got {new_value.shape}")
        if old_value.dtype != new_value.dtype:
            raise ValueError(f"Expected parameter {name} to have dtype {old_value.dtype}, got {new_value.dtype}")
    elif type(old_value) is not type(new_value):
        raise TypeError(f"Expected parameter of type {type(old_value)}, got {type(new_value)}")


def load_parameters[M: eqx.Module](
    selector: Callable[[M], Iterable[PyTree]],
    module: M,
    new_values: Iterable[PyTree],
) -> M:
    old_values = list(selector(module))
    new_values = list(new_values)
    casted_new_values = []
    for old_value, new_value in zip(old_values, new_values, strict=True):
        _check_compatible(old_value, new_value, module)
        if isinstance(old_value, Array) and isinstance(new_value, Array):
            casted_new_values.append(new_value.astype(old_value.dtype))
        else:
            casted_new_values.append(new_value)
    return eqx.tree_at(selector, module, casted_new_values, is_leaf=lambda x: x is None)



import json
from dataclasses import dataclass
from pathlib import Path
from typing import ClassVar

import cattrs
from tokenizers import AddedToken

__all__ = ["HFAddedToken", "HFTokenizerConfig"]


@dataclass(frozen=True)
class HFAddedToken:
    content: str
    single_word: bool
    normalized: bool
    special: bool
    lstrip: bool
    rstrip: bool

    def to_added_token(self) -> AddedToken:
        return AddedToken(
            self.content,
            single_word=self.single_word,
            normalized=self.normalized,
            special=self.special,
            lstrip=self.lstrip,
            rstrip=self.rstrip,
        )


@dataclass(frozen=True)
class HFTokenizerConfig:
    _converter: ClassVar[cattrs.Converter] = cattrs.Converter()
    _converter.register_structure_hook(int | list[int], lambda v, _: v)

    # ---------- core identity ----------
    tokenizer_class: str | None = None
    model_max_length: int | None = None
    padding_side: str | None = None  # "left" | "right"
    truncation_side: str | None = None  # "left" | "right"
    legacy: bool | None = None
    use_fast: bool | None = None
    clean_up_tokenization_spaces: bool | None = None

    # ---------- behaviour flags ----------
    add_bos_token: bool | None = None
    add_eos_token: bool | None = None
    add_prefix_space: bool | None = None
    use_default_system_prompt: bool | None = None
    spaces_between_special_tokens: bool | None = None
    do_lower_case: bool | None = None

    # ---------- special tokens ----------
    bos_token: str | None = None
    eos_token: str | None = None
    unk_token: str | None = None
    pad_token: str | None = None
    sep_token: str | None = None
    cls_token: str | None = None
    mask_token: str | None = None
    added_tokens_decoder: dict[str, HFAddedToken] | None = None

    # ---------- chat / SentencePiece ----------
    chat_template: str | None = None
    sp_model_kwargs: dict | None = None

    # ---------- extras ----------
    language: str | None = None
    task: str | None = None

    def added_tokens(self) -> list[AddedToken]:
        if self.added_tokens_decoder is None:
            return []
        return [token.to_added_token() for token in self.added_tokens_decoder.values()]

    @classmethod
    def from_json(cls, json_path: Path | str) -> "HFTokenizerConfig":
        json_path = Path(json_path)
        with open(json_path) as f:
            config = json.load(f)
        return cls._converter.structure(config, cls)



from collections.abc import Mapping, Sequence
from dataclasses import dataclass, replace
from typing import Self

import equinox as eqx
import jax
from jax import vmap
from jaxtyping import Array, DTypeLike, Float, Int, PRNGKeyArray

from lalamo.common import ParameterTree
from lalamo.modules.token_mixers import AttentionConfig
from lalamo.modules.utils import vmap_twice

from .common import ForwardPassMode, LalamoModule, PositionalEmbeddingSelector
from .normalization import Normalization, NormalizationConfig
from .rope import PositionalEmbeddings, RoPE, RoPEConfig
from .token_mixers import State
from .transformer_layer import (
    TransformerLayer,
    TransformerLayerConfig,
    TransformerLayerForwardPassConfig,
    TransformerLayerResult,
)

__all__ = [
    "Transformer",
    "TransformerConfig",
    "TransformerResult",
]


type TransformerForwardPassConfig = TransformerLayerForwardPassConfig


class TransformerResult(eqx.Module):
    outputs: Float[Array, "batch suffix_tokens channels"]
    updated_state: State | None = None
    layer_results: tuple[TransformerLayerResult, ...] | None = None
    global_positional_embeddings: PositionalEmbeddings | None = None
    local_positional_embeddings: PositionalEmbeddings | None = None

    def export(self) -> ParameterTree:
        result: dict[str, ParameterTree | Array] = dict(
            outputs=self.outputs,
        )
        if self.updated_state is not None:
            result["updated_state"] = [state_layer.export() for state_layer in self.updated_state]
        if self.layer_results is not None:
            result["layer_results"] = [layer_result.export() for layer_result in self.layer_results]
        if self.global_positional_embeddings is not None:
            result["global_positional_embeddings"] = self.global_positional_embeddings.export()
        if self.local_positional_embeddings is not None:
            result["local_positional_embeddings"] = self.local_positional_embeddings.export()
        return result


@dataclass(frozen=True)
class TransformerConfig:
    global_rope_config: RoPEConfig | None
    local_rope_config: RoPEConfig | None
    layer_configs: tuple[TransformerLayerConfig, ...]
    output_norm_config: NormalizationConfig
    model_dim: int
    hidden_dim: int
    context_length: int

    def random_init(self, *, key: PRNGKeyArray) -> "Transformer":
        rope_dims = (layer.rope_dim for layer in self.layer_configs if layer.rope_dim is not None)
        rope_dim = next(rope_dims, None)
        assert all(d == rope_dim for d in rope_dims)

        if self.global_rope_config:
            assert rope_dim is not None

            global_rope = self.global_rope_config.init(
                head_dim=rope_dim,
                num_timesteps=self.context_length,
            )
        else:
            global_rope = None

        if self.local_rope_config:
            assert rope_dim is not None

            max_sliding_window_size = max(
                layer_config.mixer_config.sliding_window_size or 0
                for layer_config in self.layer_configs
                if isinstance(layer_config.mixer_config, AttentionConfig)
            )

            local_rope = self.local_rope_config.init(
                head_dim=rope_dim,
                num_timesteps=max(max_sliding_window_size, self.context_length),
            )
        else:
            local_rope = None

        layers_keys = jax.random.split(key, num=len(self.layer_configs))
        layers = tuple(
            layer_config.random_init(
                model_dim=self.model_dim,
                hidden_dim=self.hidden_dim,
                key=layer_key,
            )
            for layer_key, layer_config in zip(layers_keys, self.layer_configs, strict=True)
        )
        output_norm = self.output_norm_config.init(self.model_dim)

        return Transformer(
            config=self,
            global_rope=global_rope,
            local_rope=local_rope,
            layers=layers,
            output_norm=output_norm,
        )

    def empty(self) -> "Transformer":
        rope_dims = (layer.rope_dim for layer in self.layer_configs if layer.rope_dim is not None)
        rope_dim = next(rope_dims, None)
        assert all(d == rope_dim for d in rope_dims)

        if self.global_rope_config:
            assert rope_dim is not None

            global_rope = self.global_rope_config.init(
                head_dim=rope_dim,
                num_timesteps=self.context_length,
            )
        else:
            global_rope = None

        if self.local_rope_config:
            assert rope_dim is not None

            local_rope = self.local_rope_config.init(
                head_dim=rope_dim,
                num_timesteps=self.context_length,
            )
        else:
            local_rope = None

        layers = tuple(
            layer_config.empty(
                model_dim=self.model_dim,
                hidden_dim=self.hidden_dim,
            )
            for layer_config in self.layer_configs
        )
        output_norm = self.output_norm_config.empty(self.model_dim)

        return Transformer(
            config=self,
            global_rope=global_rope,
            local_rope=local_rope,
            layers=layers,
            output_norm=output_norm,
        )


class Transformer(LalamoModule[TransformerConfig]):
    global_rope: RoPE | None
    local_rope: RoPE | None
    layers: tuple[TransformerLayer, ...]
    output_norm: Normalization

    @property
    def activation_precision(self) -> DTypeLike:
        return self.layers[0].activation_precision

    @eqx.filter_jit
    def __call__(
        self,
        inner_features: Float[Array, "batch suffix_tokens channels"],
        token_positions: Int[Array, "batch suffix_tokens"],
        state: State | None,
        return_updated_state: bool,
        return_layer_results: bool,
        return_positional_embeddings: bool,
        lengths_without_padding: Int[Array, " batch"] | None,
        forward_pass_mode: ForwardPassMode,
        forward_pass_config: TransformerForwardPassConfig | None,
    ) -> TransformerResult:
        if inner_features.ndim != 3:
            raise ValueError(
                f"inner_features must be a 3D array of size (batch_size, sequence_length, hidden_dim), got {inner_features.shape}",
            )
        if token_positions.ndim != 2:
            raise ValueError(
                "token_positions must be a 2D array of size (batch_size, sequence_length),"
                f" got {token_positions.shape}",
            )

        maybe_state = state or ([None] * len(self.layers))

        if self.global_rope is not None:
            global_positional_embeddings = vmap(self.global_rope)(token_positions)
        else:
            global_positional_embeddings = None
        if self.local_rope is not None:
            local_positional_embeddings = vmap(self.local_rope)(token_positions)
        else:
            local_positional_embeddings = global_positional_embeddings

        updated_state_layers = []
        layer_results = []

        for layer, state_layer in zip(self.layers, maybe_state, strict=True):
            match layer.positional_embedding_selector:
                case PositionalEmbeddingSelector.LOCAL:
                    positional_embeddings_to_use = local_positional_embeddings
                case PositionalEmbeddingSelector.GLOBAL:
                    positional_embeddings_to_use = global_positional_embeddings
                case PositionalEmbeddingSelector.NONE:
                    positional_embeddings_to_use = None

            layer_result = layer(
                inner_features,
                positional_embeddings_to_use,
                state=state_layer,
                return_updated_state=return_updated_state,
                return_activation_trace=return_layer_results,
                lengths_without_padding=lengths_without_padding,
                forward_pass_mode=forward_pass_mode,
                forward_pass_config=forward_pass_config,
            )
            inner_features = layer_result.outputs
            layer_results.append(layer_result)
            updated_state_layers.append(layer_result.updated_state)

        normalized_outputs = vmap_twice(self.output_norm)(inner_features)

        return TransformerResult(
            outputs=normalized_outputs,
            updated_state=(State(updated_state_layers) if return_updated_state else None),
            layer_results=tuple(layer_results) if return_layer_results else None,
            global_positional_embeddings=(global_positional_embeddings if return_positional_embeddings else None),
            local_positional_embeddings=(local_positional_embeddings if return_positional_embeddings else None),
        )

    def init_static_state(self, batch_size: int, capacity: int) -> State:
        return State(layer.init_static_state(batch_size, capacity) for layer in self.layers)

    def export_weights(self) -> ParameterTree:
        result = dict(
            layers=[layer.export_weights() for layer in self.layers],
            output_norm=self.output_norm.export_weights(),
        )
        if self.global_rope:
            result["global_rope"] = self.global_rope.export_weights()
        if self.local_rope:
            result["local_rope"] = self.local_rope.export_weights()
        return result

    def import_weights(
        self,
        weights: ParameterTree[Array],
    ) -> Self:
        assert isinstance(weights, Mapping)
        assert isinstance(weights["layers"], Sequence)
        assert isinstance(weights["output_norm"], Mapping)

        if self.global_rope:
            assert isinstance(weights["global_rope"], Mapping)
            global_rope = self.global_rope.import_weights(weights["global_rope"])
        else:
            global_rope = None

        if self.local_rope:
            assert isinstance(weights["local_rope"], Mapping)
            local_rope = self.local_rope.import_weights(weights["local_rope"])
        else:
            local_rope = None

        layers = []
        for layer, layer_weights in zip(self.layers, weights["layers"], strict=True):
            assert isinstance(layer_weights, Mapping)
            layers.append(layer.import_weights(layer_weights))

        return replace(
            self,
            global_rope=global_rope,
            layers=tuple(layers),
            output_norm=self.output_norm.import_weights(weights["output_norm"]),
            local_rope=local_rope,
        )



from abc import abstractmethod
from collections.abc import Iterable
from typing import Self


class Speculator:
    @abstractmethod
    def train(self, token_ids: Iterable[int], token_logits: Iterable[dict[int, float]]) -> None:
        raise NotImplementedError

    @abstractmethod
    def probs(self, seq: Iterable[int]) -> dict[int, float]:
        raise NotImplementedError

    @abstractmethod
    def serialize(self) -> bytes:
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def deserialize(cls, blob: bytes) -> Self:
        raise NotImplementedError



from lalamo.model_import.decoder_configs import HFQwen2Config

from .common import (
    ModelSpec,
)

__all__ = ["DEEPSEEK_MODELS"]

DEEPSEEK_MODELS = [
    ModelSpec(
        vendor="DeepSeek",
        family="R1-Distill-Qwen",
        name="R1-Distill-Qwen-1.5B",
        size="1.5B",
        quantization=None,
        repo="deepseek-ai/DeepSeek-R1-Distill-Qwen-1.5B",
        config_type=HFQwen2Config,
    ),
]



from collections.abc import (
    Callable,
    Collection,
    Iterable,
    Iterator,
    KeysView,
    Mapping,
    MappingView,
    Sequence,
    ValuesView,
)
from contextlib import contextmanager
from dataclasses import dataclass
from pathlib import Path
from typing import overload

import einops
import jax.numpy as jnp
from jaxtyping import Array
from safetensors import safe_open

__all__ = [
    "MapDictValues",
    "MapSequence",
    "jax_uint4_to_packed_uint8",
    "open_safetensors",
    "process_chat_template",
]


@dataclass(frozen=True)
class LazyDict[K, V](Mapping[K, V]):
    stored_keys: set[K]
    getter: Callable[[K], V]

    def __getitem__(self, key: K) -> V:
        if key not in self.stored_keys:
            raise KeyError(key)
        return self.getter(key)

    def __iter__(self) -> Iterator[K]:
        return iter(self.stored_keys)

    def __len__(self) -> int:
        return len(self.stored_keys)


@contextmanager
def open_safetensors(filename: Path | str) -> Iterator[tuple[Mapping[str, Array], Mapping[str, str]]]:
    with safe_open(filename, framework="flax") as safetensors_nonsense:
        yield (
            LazyDict(set(safetensors_nonsense.keys()), safetensors_nonsense.get_tensor),
            safetensors_nonsense.metadata(),
        )


@dataclass(frozen=True)
class MapIterable[OldT, NewT](Iterable[NewT]):
    map_func: Callable[[OldT], NewT]
    collection: Iterable[OldT]

    def __iter__(self) -> Iterator[NewT]:
        return map(self.map_func, self.collection)


@dataclass(frozen=True)
class MapCollection[OldT, NewT](MapIterable[OldT, NewT], Collection[NewT]):
    collection: Collection[OldT]

    def __contains__(self, item: object) -> bool:
        return any(self.map_func(elem) == item for elem in self.collection)

    def __len__(self) -> int:
        return len(self.collection)


@dataclass(frozen=True)
class MapSequence[OldT, NewT](MapCollection[OldT, NewT], Sequence[NewT]):
    collection: Sequence[OldT]

    @overload
    def __getitem__(self, index: int) -> NewT: ...

    @overload
    def __getitem__(self, index: slice) -> list[NewT]: ...

    def __getitem__(self, index):
        if isinstance(index, slice):
            return [self.map_func(item) for item in self.collection[index]]
        return self.map_func(self.collection[index])


@dataclass(frozen=True)
class MapView[OldT, NewT](MapCollection[OldT, NewT], MappingView):
    pass


@dataclass(frozen=True)
class MapValuesView[OldT, NewT](MapCollection[OldT, NewT], ValuesView):
    pass


@dataclass(frozen=True)
class MapDictValues[K, OldV, NewV](Mapping[K, NewV]):
    value_map: Callable[[OldV], NewV]
    collection: Mapping[K, OldV]

    def __getitem__(self, key: K) -> NewV:
        return self.value_map(self.collection[key])

    def __iter__(self) -> Iterator[K]:
        return iter(self.collection)

    def keys(self) -> KeysView[K]:
        return self.collection.keys()

    def values(self) -> MapValuesView[OldV, NewV]:
        return MapValuesView(self.value_map, self.collection.values())

    def __len__(self) -> int:
        return len(self.collection)


def jax_uint4_to_packed_uint8(array: Array) -> Array:
    if array.dtype != jnp.uint4:
        raise ValueError(f"Input array must have dtype jnp.uint4, but got {array.dtype}")

    if not array.shape:
        raise ValueError("Input array cannot be a scalar and must have at least one dimension.")

    *_, last_dim = array.shape
    if last_dim % 2 != 0:
        raise ValueError(f"The last dimension of the input array must be even, but got shape {array.shape}")

    low_nibbles, high_nibbles = einops.rearrange(
        array.astype(jnp.uint8),
        "... (dim_half two) -> two ... dim_half",
        two=2,
    )

    packed = (high_nibbles << 4) | low_nibbles

    return packed.astype(jnp.uint8)


def jax_uint8_to_unpacked_uint4(array: Array) -> Array:
    if array.dtype != jnp.uint8:
        raise ValueError(f"Input array must have dtype jnp.uint8, but got {array.dtype}")

    if not array.shape:
        raise ValueError("Input array cannot be a scalar and must have at least one dimension.")

    low_nibbles = array & 0x0F
    high_nibbles = (array >> 4) & 0xF

    unpacked = einops.rearrange(
        jnp.stack([low_nibbles, high_nibbles]),
        "two ... dim_half -> ... (dim_half two)",
        two=2,
    )

    return unpacked.astype(jnp.uint4)


def process_chat_template(template: str) -> str:
    template = template.replace("{% generation %}", "")
    template = template.replace("{%- endgeneration -%}", "")
    return template



from lalamo.modules.common import register_config_union

from .attention import Attention, AttentionConfig, AttentionResult
from .common import TokenMixerBase, TokenMixerResult
from .mamba import Mamba2, Mamba2Config, Mamba2Result, SeparableCausalConv, SeparableCausalConvConfig
from .short_conv import ShortConv, ShortConvConfig, ShortConvResult
from .state import (
    DynamicKVCacheLayer,
    KVCacheLayer,
    Mamba2StateLayer,
    ShortConvStateLayer,
    State,
    StateLayerBase,
    StaticKVCacheLayer,
)

TokenMixerConfig = AttentionConfig | Mamba2Config | ShortConvConfig

register_config_union(TokenMixerConfig)  # type: ignore (pyright bug)

__all__ = [
    "Attention",
    "AttentionConfig",
    "AttentionResult",
    "DynamicKVCacheLayer",
    "KVCacheLayer",
    "Mamba2",
    "Mamba2Config",
    "Mamba2Result",
    "Mamba2StateLayer",
    "SeparableCausalConv",
    "SeparableCausalConvConfig",
    "ShortConv",
    "ShortConvConfig",
    "ShortConvResult",
    "ShortConvStateLayer",
    "State",
    "StateLayerBase",
    "StaticKVCacheLayer",
    "TokenMixerBase",
    "TokenMixerConfig",
    "TokenMixerResult",
]



from collections.abc import Mapping
from dataclasses import dataclass, replace
from functools import partial
from typing import Self

import equinox as eqx
import jax
import jax.numpy as jnp
from jax import vmap
from jaxtyping import Array, DTypeLike, Float, Int, PRNGKeyArray

from lalamo.common import ParameterTree

from .common import ForwardPassMode, LalamoModule, PositionalEmbeddingSelector
from .mlp import MLPBase, MLPConfig, MLPForwardPassConfig
from .normalization import Normalization, NormalizationConfig
from .rope import PositionalEmbeddings
from .token_mixers import KVCacheLayer, StateLayerBase, StaticKVCacheLayer, TokenMixerBase, TokenMixerConfig
from .utils import vmap_twice

__all__ = [
    "TransformerLayer",
    "TransformerLayerActivationTrace",
    "TransformerLayerConfig",
    "TransformerLayerForwardPassConfig",
    "TransformerLayerResult",
]


type TransformerLayerForwardPassConfig = MLPForwardPassConfig


class TransformerLayerActivationTrace(eqx.Module):
    inputs: Float[Array, "batch suffix_tokens channels"]
    positional_embeddings: PositionalEmbeddings | None
    state: StateLayerBase | None

    mlp_inputs: Float[Array, "batch suffix_tokens channels"]
    pre_mixer_norm: Float[Array, "batch suffix_tokens channels"]
    mixer: Float[Array, "batch suffix_tokens channels"]
    post_mixer_norm: Float[Array, "batch suffix_tokens channels"] | None
    pre_mlp_norm: Float[Array, "batch suffix_tokens channels"]
    mlp: Float[Array, "batch suffix_tokens channels"]
    post_mlp_norm: Float[Array, "batch suffix_tokens channels"] | None

    def export(self) -> ParameterTree:
        result: dict[str, ParameterTree | Array] = dict(
            inputs=self.inputs,
            mlp_inputs=self.mlp_inputs,
            pre_mixer_norm=self.pre_mixer_norm,
            mixer=self.mixer,
            pre_mlp_norm=self.pre_mlp_norm,
            mlp=self.mlp,
        )
        if self.positional_embeddings is not None:
            result["positional_embeddings"] = self.positional_embeddings.export()
        if self.state is not None:
            result["state"] = self.state.export()
        if self.post_mixer_norm is not None:
            result["post_mixer_norm"] = self.post_mixer_norm
        if self.post_mlp_norm is not None:
            result["post_mlp_norm"] = self.post_mlp_norm
        return result


class TransformerLayerResult(eqx.Module):
    outputs: Float[Array, "batch tokens channels"]
    updated_state: KVCacheLayer | None
    activation_trace: TransformerLayerActivationTrace | None

    def export(self) -> ParameterTree:
        result: dict[str, ParameterTree | Array] = dict(
            outputs=self.outputs,
        )
        if self.updated_state is not None:
            result["updated_state"] = self.updated_state.export()
        if self.activation_trace is not None:
            result["activation_trace"] = self.activation_trace.export()
        return result


@dataclass(frozen=True)
class TransformerLayerConfig:
    pre_mixer_norm_config: NormalizationConfig | None
    mixer_config: TokenMixerConfig
    post_mixer_norm_config: NormalizationConfig | None
    pre_mlp_norm_config: NormalizationConfig
    mlp_config: MLPConfig
    post_mlp_norm_config: NormalizationConfig | None

    @property
    def rope_dim(self) -> int | None:
        return self.mixer_config.rope_dim

    def random_init(
        self,
        model_dim: int,
        hidden_dim: int,
        *,
        key: PRNGKeyArray,
    ) -> "TransformerLayer":
        attention_key, mlp_key = jax.random.split(key)
        if self.pre_mixer_norm_config is not None:
            pre_mixer_norm = self.pre_mixer_norm_config.init(model_dim)
        else:
            pre_mixer_norm = None
        mixer = self.mixer_config.random_init(
            model_dim=model_dim,
            key=attention_key,
        )
        if self.post_mixer_norm_config is not None:
            post_mixer_norm = self.post_mixer_norm_config.init(model_dim)
        else:
            post_mixer_norm = None
        pre_mlp_norm = self.pre_mlp_norm_config.init(model_dim)
        mlp = self.mlp_config.random_init(model_dim, hidden_dim, key=mlp_key)
        if self.post_mlp_norm_config is not None:
            post_mlp_norm = self.post_mlp_norm_config.init(model_dim)
        else:
            post_mlp_norm = None
        return TransformerLayer(
            config=self,
            pre_mixer_norm=pre_mixer_norm,
            mixer=mixer,
            post_mixer_norm=post_mixer_norm,
            pre_mlp_norm=pre_mlp_norm,
            mlp=mlp,
            post_mlp_norm=post_mlp_norm,
        )

    def empty(
        self,
        model_dim: int,
        hidden_dim: int,
    ) -> "TransformerLayer":
        if self.pre_mixer_norm_config is not None:
            pre_mixer_norm = self.pre_mixer_norm_config.empty(model_dim)
        else:
            pre_mixer_norm = None
        attention = self.mixer_config.empty(
            model_dim=model_dim,
        )
        if self.post_mixer_norm_config is not None:
            post_mixer_norm = self.post_mixer_norm_config.empty(model_dim)
        else:
            post_mixer_norm = None
        pre_mlp_norm = self.pre_mlp_norm_config.empty(model_dim)
        mlp = self.mlp_config.empty(model_dim, hidden_dim)
        if self.post_mlp_norm_config is not None:
            post_mlp_norm = self.post_mlp_norm_config.empty(model_dim)
        else:
            post_mlp_norm = None
        return TransformerLayer(
            config=self,
            pre_mixer_norm=pre_mixer_norm,
            mixer=attention,
            post_mixer_norm=post_mixer_norm,
            pre_mlp_norm=pre_mlp_norm,
            mlp=mlp,
            post_mlp_norm=post_mlp_norm,
        )


class TransformerLayer(LalamoModule[TransformerLayerConfig]):
    pre_mixer_norm: Normalization | None
    mixer: TokenMixerBase
    post_mixer_norm: Normalization | None
    pre_mlp_norm: Normalization
    mlp: MLPBase
    post_mlp_norm: Normalization | None

    @property
    def activation_precision(self) -> DTypeLike:
        return self.mixer.activation_precision

    @property
    def positional_embedding_selector(self) -> PositionalEmbeddingSelector:
        return self.mixer.positional_embedding_selector

    def __post_init__(self) -> None:
        model_dim = self.pre_mixer_norm.input_dim if self.pre_mixer_norm is not None else self.mixer.model_dim
        if self.mixer.model_dim != model_dim:
            raise ValueError(
                f"Attention model dim {self.mixer.model_dim} does not match"
                f" the first normalization layer dim {model_dim}",
            )
        if self.post_mixer_norm is not None and self.post_mixer_norm.input_dim != model_dim:
            raise ValueError(
                f"Post mixer normalization dim {self.post_mixer_norm.input_dim} does not match"
                f" the first normalization layer dim {model_dim}",
            )
        if self.pre_mlp_norm.input_dim != model_dim:
            raise ValueError(
                f"Pre MLP normalization dim {self.pre_mlp_norm.input_dim} does not match"
                f" the first normalization layer dim {model_dim}",
            )
        if self.mlp.model_dim != model_dim:
            raise ValueError(
                f"MLP up projection dim {self.mlp.model_dim} does not match"
                f" the first normalization layer dim {model_dim}",
            )

    @eqx.filter_jit
    def __call__(
        self,
        inputs: Float[Array, "batch suffix_tokens channels"],
        positional_embeddings: PositionalEmbeddings | None,
        state: StateLayerBase | None = None,
        return_updated_state: bool = False,
        return_activation_trace: bool = False,
        lengths_without_padding: Int[Array, " batch"] | None = None,
        forward_pass_mode: ForwardPassMode = ForwardPassMode.MULTI_TOKEN,
        forward_pass_config: TransformerLayerForwardPassConfig | None = None,
    ) -> TransformerLayerResult:
        if inputs.ndim != 3:
            raise ValueError(
                f"Inputs to decoder layers must be a 3D arrays of size (batch_size, sequence_length, hidden_dim),"
                f" got {inputs.shape}",
            )
        if self.pre_mixer_norm is not None:
            normalized_mixer_inputs = vmap_twice(self.pre_mixer_norm)(inputs)
        else:
            normalized_mixer_inputs = inputs

        batched_mixer_fn = vmap(
            partial(self.mixer, return_updated_state=return_updated_state or return_activation_trace),
        )
        mixer_outputs, updated_state = batched_mixer_fn(
            normalized_mixer_inputs,
            positional_embeddings,
            state=state,
            length_without_padding=lengths_without_padding,
        )
        if self.post_mixer_norm is not None:
            normalized_mixer_outputs = vmap_twice(self.post_mixer_norm)(mixer_outputs)
            mlp_inputs = inputs + normalized_mixer_outputs
        else:
            normalized_mixer_outputs = None
            mlp_inputs = inputs + mixer_outputs

        normalized_mlp_inputs = vmap_twice(self.pre_mlp_norm)(mlp_inputs)
        mlp_outputs = self.mlp(
            normalized_mlp_inputs,
            forward_pass_mode=forward_pass_mode,
            forward_pass_config=forward_pass_config,
        )
        if self.post_mlp_norm is not None:
            normalized_mlp_outputs = vmap_twice(self.post_mlp_norm)(mlp_outputs)
            outputs = mlp_inputs + normalized_mlp_outputs
        else:
            normalized_mlp_outputs = None
            outputs = mlp_inputs + mlp_outputs

        if return_activation_trace:
            activation_trace = TransformerLayerActivationTrace(
                inputs=inputs,
                positional_embeddings=positional_embeddings,
                state=state,
                pre_mixer_norm=normalized_mixer_inputs,
                mixer=mixer_outputs,
                post_mixer_norm=normalized_mixer_outputs,
                mlp_inputs=mlp_inputs,
                pre_mlp_norm=normalized_mlp_inputs,
                mlp=mlp_outputs,
                post_mlp_norm=normalized_mlp_outputs,
            )
        else:
            activation_trace = None

        return TransformerLayerResult(
            outputs=outputs,
            updated_state=updated_state,
            activation_trace=activation_trace,
        )

    def init_static_state(self, batch_size: int, capacity: int) -> StaticKVCacheLayer:
        return jax.tree.map(
            lambda array: jnp.repeat(array[None, ...], batch_size, axis=0),
            self.mixer.init_static_state(capacity),
        )

    def export_weights(self) -> ParameterTree:
        result = dict(
            mixer=self.mixer.export_weights(),
            pre_mlp_norm=self.pre_mlp_norm.export_weights(),
            mlp=self.mlp.export_weights(),
        )
        if self.pre_mixer_norm is not None:
            result["pre_mixer_norm"] = self.pre_mixer_norm.export_weights()
        if self.post_mixer_norm is not None:
            result["post_mixer_norm"] = self.post_mixer_norm.export_weights()
        if self.post_mlp_norm is not None:
            result["post_mlp_norm"] = self.post_mlp_norm.export_weights()
        return result

    def import_weights(
        self,
        weights: ParameterTree[Array],
    ) -> Self:
        assert isinstance(weights, Mapping)
        assert isinstance(weights["mixer"], Mapping)
        assert isinstance(weights["mlp"], Mapping)
        assert isinstance(weights["pre_mlp_norm"], Mapping)

        if self.post_mixer_norm is not None:
            assert isinstance(weights["post_mixer_norm"], Mapping)
            post_mixer_norm = self.post_mixer_norm.import_weights(
                weights["post_mixer_norm"],
            )
        else:
            post_mixer_norm = None
        if self.post_mlp_norm is not None:
            assert isinstance(weights["post_mlp_norm"], Mapping)
            post_mlp_norm = self.post_mlp_norm.import_weights(weights["post_mlp_norm"])
        else:
            post_mlp_norm = None
        if self.pre_mixer_norm is not None:
            assert isinstance(weights["pre_mixer_norm"], Mapping)
            pre_mixer_norm = self.pre_mixer_norm.import_weights(weights["pre_mixer_norm"])
        else:
            pre_mixer_norm = None
        return replace(
            self,
            pre_mixer_norm=pre_mixer_norm,
            mixer=self.mixer.import_weights(weights["mixer"]),
            post_mixer_norm=post_mixer_norm,
            pre_mlp_norm=self.pre_mlp_norm.import_weights(weights["pre_mlp_norm"]),
            mlp=self.mlp.import_weights(weights["mlp"]),
            post_mlp_norm=post_mlp_norm,
        )



from .activations import GELU, Activation, Identity, SiLU
from .classifier import Classifier, ClassifierConfig
from .common import (
    ForwardPassMode,
    LalamoModule,
    PositionalEmbeddingSelector,
    config_converter,
)
from .decoder import (
    Decoder,
    DecoderActivationTrace,
    DecoderConfig,
    DecoderForwardPassConfig,
    DecoderResult,
)
from .embedding import (
    EmbeddingBase,
    EmbeddingConfig,
    MLXQuantizedTiedEmbedding,
    MLXQuantizedTiedEmbeddingConfig,
    MLXSemiQuantizedUntiedEmbedding,
    MLXSemiQuantizedUntiedEmbeddingConfig,
    QuantizedTiedEmbedding,
    QuantizedTiedEmbeddingConfig,
    TiedEmbedding,
    TiedEmbeddingConfig,
    UntiedEmbedding,
    UntiedEmbeddingConfig,
)
from .linear import (
    FullPrecisionLinear,
    FullPrecisionLinearConfig,
    GroupQuantizedLinear,
    GroupQuantizedLinearConfig,
    LinearBase,
    LinearConfig,
    MLXQuantizedLinear,
    MLXQuantizedLinearConfig,
    QLoRALinear,
    QLoRALinearConfig,
)
from .mlp import (
    DenseMLP,
    DenseMLPConfig,
    MixtureOfExperts,
    MixtureOfExpertsConfig,
    MLPBase,
    MLPConfig,
    MLPForwardPassConfig,
    RoutingFunction,
    SoftmaxRouting,
)
from .normalization import Normalization, NormalizationConfig, UpcastMode
from .rope import (
    LinearScalingRoPEConfig,
    LlamaRoPEConfig,
    PositionalEmbeddings,
    RoPE,
    RoPEConfig,
    UnscaledRoPEConfig,
    YARNRoPEConfig,
)
from .token_mixers import (
    Attention,
    AttentionConfig,
    DynamicKVCacheLayer,
    KVCacheLayer,
    Mamba2,
    Mamba2Config,
    SeparableCausalConv,
    SeparableCausalConvConfig,
    ShortConv,
    ShortConvConfig,
    ShortConvStateLayer,
    State,
    StaticKVCacheLayer,
)
from .transformer import Transformer, TransformerConfig
from .transformer_layer import (
    TransformerLayer,
    TransformerLayerActivationTrace,
    TransformerLayerConfig,
    TransformerLayerForwardPassConfig,
    TransformerLayerResult,
)

__all__ = [
    "GELU",
    "Activation",
    "Attention",
    "AttentionConfig",
    "Classifier",
    "ClassifierConfig",
    "Decoder",
    "DecoderActivationTrace",
    "DecoderConfig",
    "DecoderForwardPassConfig",
    "DecoderResult",
    "DenseMLP",
    "DenseMLPConfig",
    "DynamicKVCacheLayer",
    "EmbeddingBase",
    "EmbeddingConfig",
    "ForwardPassMode",
    "FullPrecisionLinear",
    "FullPrecisionLinearConfig",
    "GroupQuantizedLinear",
    "GroupQuantizedLinearConfig",
    "Identity",
    "KVCacheLayer",
    "LalamoModule",
    "LinearBase",
    "LinearConfig",
    "LinearScalingRoPEConfig",
    "LlamaRoPEConfig",
    "MLPBase",
    "MLPConfig",
    "MLPForwardPassConfig",
    "MLXQuantizedLinear",
    "MLXQuantizedLinearConfig",
    "MLXQuantizedTiedEmbedding",
    "MLXQuantizedTiedEmbeddingConfig",
    "MLXSemiQuantizedUntiedEmbedding",
    "MLXSemiQuantizedUntiedEmbeddingConfig",
    "Mamba2",
    "Mamba2Config",
    "MixtureOfExperts",
    "MixtureOfExpertsConfig",
    "Normalization",
    "NormalizationConfig",
    "PositionalEmbeddingSelector",
    "PositionalEmbeddings",
    "QLoRALinear",
    "QLoRALinearConfig",
    "QuantizedTiedEmbedding",
    "QuantizedTiedEmbeddingConfig",
    "RoPE",
    "RoPEConfig",
    "RoutingFunction",
    "SeparableCausalConv",
    "SeparableCausalConvConfig",
    "ShortConv",
    "ShortConvConfig",
    "ShortConvStateLayer",
    "SiLU",
    "SoftmaxRouting",
    "State",
    "StaticKVCacheLayer",
    "TiedEmbedding",
    "TiedEmbeddingConfig",
    "Transformer",
    "TransformerConfig",
    "TransformerLayer",
    "TransformerLayerActivationTrace",
    "TransformerLayerConfig",
    "TransformerLayerForwardPassConfig",
    "TransformerLayerResult",
    "UnscaledRoPEConfig",
    "UntiedEmbedding",
    "UntiedEmbeddingConfig",
    "UpcastMode",
    "YARNRoPEConfig",
    "config_converter",
]



from enum import Enum

from jax import numpy as jnp
from jaxtyping import Array, DTypeLike, Float

__all__ = ["QuantizationMode", "quantize_weights"]


class QuantizationMode(Enum):
    UINT4 = "uint4"
    INT8 = "int8"
    UINT8 = "uint8"

    @classmethod
    def from_num_bits(cls, num_bits: int) -> "QuantizationMode":
        bit_to_mode = {
            4: cls.UINT4,
            8: cls.UINT8,
        }
        if num_bits not in bit_to_mode:
            raise ValueError(f"No quantization mode defined for {num_bits} bits")
        return bit_to_mode[num_bits]

    @property
    def range(self) -> tuple[int, int]:
        return MODE_TO_RANGE[self]

    @property
    def dtype(self) -> DTypeLike:
        value_to_dtype = {
            QuantizationMode.UINT4: jnp.uint4,
            QuantizationMode.INT8: jnp.int8,
            QuantizationMode.UINT8: jnp.uint8,
        }
        return value_to_dtype[self]

    @property
    def bits(self) -> int:
        value_to_bits = {
            QuantizationMode.UINT4: 4,
            QuantizationMode.INT8: 8,
            QuantizationMode.UINT8: 8,
        }
        return value_to_bits[self]

    def __str__(self) -> str:
        return self.value


MODE_TO_RANGE = {
    QuantizationMode.UINT4: (0, 15),
    QuantizationMode.INT8: (-128, 127),
    QuantizationMode.UINT8: (0, 255),
}


def quantize_weights(x: Float[Array, "..."], mode: QuantizationMode) -> Float[Array, "..."]:
    range_min, range_max = MODE_TO_RANGE[mode]
    return jnp.clip(jnp.round(x), range_min, range_max)


def dynamically_quantize_activations(
    x: Float[Array, " channels"],
    mode: QuantizationMode,
) -> Float[Array, " channels"]:
    # Reference implementation: https://github.com/pytorch/pytorch/blob/2ccbacfa24cae724ec1ea3bc7de189e5bf948d46/torch/ao/quantization/fx/_decomposed.py#L790
    range_min, range_max = mode.range
    min_val = jnp.min(x)
    max_val = jnp.max(x)
    min_val_neg = jnp.minimum(min_val, 0)
    max_val_pos = jnp.maximum(max_val, 0)

    # scale
    scale = (max_val_pos - min_val_neg) / (range_max - range_min)
    scale = jnp.maximum(scale, jnp.finfo(x.dtype).eps)

    # zero point
    descaled_min = min_val_neg / scale
    descaled_max = max_val_pos / scale
    zero_point_from_min_error = range_min + descaled_min
    zero_point_from_max_error = range_max + descaled_max
    zero_point = jnp.where(
        zero_point_from_min_error + zero_point_from_max_error > 0,
        range_min - descaled_min,
        range_max - descaled_max,
    )
    zero_point = jnp.round(jnp.clip(zero_point, range_min, range_max))

    x_normalized = x / scale + zero_point
    x_quantized = jnp.clip(jnp.round(x_normalized), range_min, range_max)

    return (x_quantized - zero_point) * scale



from collections.abc import Mapping
from dataclasses import dataclass

import jax.numpy as jnp
from jaxtyping import Array, DTypeLike

from lalamo.model_import.loaders.executorch import load_executorch
from lalamo.modules import (
    AttentionConfig,
    Decoder,
    DecoderConfig,
    DenseMLPConfig,
    LlamaRoPEConfig,
    NormalizationConfig,
    QLoRALinearConfig,
    QuantizedTiedEmbeddingConfig,
    TransformerConfig,
    TransformerLayerConfig,
    UpcastMode,
)
from lalamo.modules.activations import SiLU
from lalamo.quantization import QuantizationMode

from .common import ForeignConfig

__all__ = ["ETLlamaConfig"]


# These parameters are not present in the config file, and are extracted from the executorch implementation
LOW_FREQ_FACTOR = 1.0
HIGH_FREQ_FACTOR = 4.0
OLD_CONTEXT_LENGTH = 8192
MAX_SEQUENCE_LENGTH = 8192 * 32

ROPE_SCALING_FACTOR = 32.0

EMBEDDING_QUANTIZATION_MODE = QuantizationMode.INT8
ACTIVATION_QUANTIZATION_MODE = QuantizationMode.INT8
WEIGHT_QUANTIZATION_MODE = QuantizationMode.UINT4


@dataclass(frozen=True)
class QuantizationConfig:
    group_size: int


@dataclass(frozen=True)
class LoraConfig:
    rank: int
    scale: float


@dataclass(frozen=True)
class ExecutorchConfig(ForeignConfig):
    eos_token_id: int | list[int]

    @property
    def eos_token_ids(self) -> list[int]:
        return [self.eos_token_id] if isinstance(self.eos_token_id, int) else self.eos_token_id

    @property
    def default_precision(self) -> DTypeLike:
        return jnp.bfloat16

    @classmethod
    def _load_decoder_weights(
        cls,
        model: Decoder,
        weights_dict: Mapping[str, Array],
    ) -> Decoder:
        return load_executorch(model, weights_dict)


@dataclass(frozen=True)
class ETLlamaConfig(ExecutorchConfig):
    dim: int
    n_layers: int
    n_heads: int
    n_kv_heads: int
    vocab_size: int
    ffn_dim_multiplier: float
    multiple_of: int
    norm_eps: float
    rope_theta: float
    use_scaled_rope: bool
    quantization_args: QuantizationConfig | None = None
    lora_args: LoraConfig | None = None

    def _find_hidden_size(self) -> int:
        # Magic formula from executorch
        size_candidate = int(8 / 3 * self.dim * self.ffn_dim_multiplier)
        return size_candidate // self.multiple_of * self.multiple_of

    def to_decoder_config(
        self,
        context_length: int | None,
        activation_precision: DTypeLike,
        accumulation_precision: DTypeLike,
        metadata_dict: Mapping[str, str],  # noqa: ARG002
    ) -> DecoderConfig:
        if self.lora_args is None:
            raise ValueError("We only support QLoRA models for now.")

        if self.quantization_args is None:
            raise ValueError("Quantization arguments are required for QLoRA models.")

        embedding_config = QuantizedTiedEmbeddingConfig(
            input_scale=None,
            logit_soft_cap=None,
            embedding_quantization_mode=EMBEDDING_QUANTIZATION_MODE,
            activation_quantization_mode=ACTIVATION_QUANTIZATION_MODE,
            activation_precision=activation_precision,
        )
        rope_config = LlamaRoPEConfig(
            precision=activation_precision,
            base=self.rope_theta,
            max_sequence_length=MAX_SEQUENCE_LENGTH,
            scaling_factor=ROPE_SCALING_FACTOR,
            original_context_length=OLD_CONTEXT_LENGTH,
            low_frequency_factor=LOW_FREQ_FACTOR,
            high_frequency_factor=HIGH_FREQ_FACTOR,
        )
        rmsnorm_config = NormalizationConfig(
            scale_precision=activation_precision,
            accumulation_precision=accumulation_precision,
            epsilon=self.norm_eps,
            scale_offset=None,
            upcast_mode=UpcastMode.ONLY_NORMALIZATION,
            subtract_mean=False,
        )
        linear_config = QLoRALinearConfig(
            group_size=self.quantization_args.group_size,
            weight_quantization_mode=WEIGHT_QUANTIZATION_MODE,
            activation_quantization_mode=ACTIVATION_QUANTIZATION_MODE,
            activation_precision=activation_precision,
            lora_rank=self.lora_args.rank,
            lora_scale=self.lora_args.scale,
        )
        attention_config = AttentionConfig(
            qkv_projection_config=linear_config,
            out_projection_config=linear_config,
            query_norm_config=None,
            key_norm_config=None,
            logit_soft_cap=None,
            has_sinks=False,
            has_qkv_biases=False,
            has_out_biases=False,
            num_heads=self.n_heads,
            num_groups=self.n_kv_heads,
            head_dim=self.dim // self.n_heads,
            is_causal=True,
            scale=None,
            sliding_window_size=None,
        )
        mlp_config = DenseMLPConfig(
            linear_config=linear_config,
            activation=SiLU(),
            has_up_biases=False,
            has_down_biases=False,
            up_clipping=None,
            gate_clipping=None,
        )
        tranformer_layer_config = TransformerLayerConfig(
            pre_mixer_norm_config=rmsnorm_config,
            mixer_config=attention_config,
            post_mixer_norm_config=None,
            pre_mlp_norm_config=rmsnorm_config,
            mlp_config=mlp_config,
            post_mlp_norm_config=None,
        )
        transformer_config = TransformerConfig(
            global_rope_config=rope_config,
            local_rope_config=None,
            layer_configs=(tranformer_layer_config,) * self.n_layers,
            output_norm_config=rmsnorm_config,
            model_dim=self.dim,
            hidden_dim=self._find_hidden_size(),
            context_length=context_length or MAX_SEQUENCE_LENGTH,
        )
        return DecoderConfig(
            embedding_config=embedding_config,
            transformer_config=transformer_config,
            vocab_size=self.vocab_size,
        )



import functools
import itertools
from collections.abc import Callable
from typing import NamedTuple

import jax
import jax.numpy as jnp

from lalamo.models import LanguageModel


def estimate_memory_from_batchsize(
    model: LanguageModel,
    max_input_length: int,
    max_output_length: int,
    num_logits_per_token: int,
    batch_size: int,
) -> int:
    memory_analysis = (
        jax.jit(
            functools.partial(
                LanguageModel.generate_tokens,
                max_output_length=max_output_length,
                num_top_logits_to_return=num_logits_per_token,
            ),
            backend="cpu", # cuda backend tries to allocate in .compile() and ooms
        )
        .lower(
            model,
            prompt_token_ids=jax.ShapeDtypeStruct((batch_size, max_input_length), jnp.int32),
            prompt_lengths_without_padding=jax.ShapeDtypeStruct((batch_size,), jnp.int32),
        )
        .compile()
        .memory_analysis()
    )

    assert hasattr(memory_analysis, "argument_size_in_bytes")
    assert hasattr(memory_analysis, "output_size_in_bytes")
    assert hasattr(memory_analysis, "temp_size_in_bytes")

    return (
        memory_analysis.argument_size_in_bytes  # type: ignore (pyright bug)
        + memory_analysis.output_size_in_bytes  # type: ignore (pyright bug)
        + memory_analysis.temp_size_in_bytes    # type: ignore (pyright bug)
    )


class EstimateBatchsizeFromMemoryEvent(NamedTuple):
    lo: int
    hi: int | None


def estimate_batchsize_from_memory(
    model: LanguageModel,
    max_input_length: int,
    max_output_length: int,
    num_logits_per_token: int,
    target_mem: int,
    progress: Callable[[EstimateBatchsizeFromMemoryEvent], None] | None = None,
) -> int:
    mem_for_bs = functools.cache(
        functools.partial(
            estimate_memory_from_batchsize,
            model,
            max_input_length,
            max_output_length,
            num_logits_per_token,
        ),
    )

    lo = 0
    hi = 0
    for candidate_exp in itertools.count():
        lo = hi
        hi = 2**candidate_exp

        if progress is not None:
            progress(EstimateBatchsizeFromMemoryEvent(lo, None))
        if target_mem < mem_for_bs(hi):
            break

    while hi - lo > 1:
        mid = (lo + hi) // 2

        if progress is not None:
            progress(EstimateBatchsizeFromMemoryEvent(lo, hi))
        if target_mem < mem_for_bs(mid):
            hi = mid
        else:
            lo = mid

    return lo



import pytest

from lalamo.model_import.decoder_configs.huggingface.llama import HFLlamaConfig
from lalamo.model_import.model_specs.common import ConfigMap, FileSpec, JSONFieldSpec, ModelSpec

DIRECT_TEMPLATE = "{% for message in messages %}{{ message.content }}{% endfor %}"


class TestConfigMapChatTemplate:
    def test_chat_template_as_string(self) -> None:
        config = ConfigMap(chat_template=DIRECT_TEMPLATE)
        assert config.chat_template == DIRECT_TEMPLATE
        assert isinstance(config.chat_template, str)

    def test_chat_template_as_file_spec(self) -> None:
        file_spec = FileSpec("chat_template.jinja")
        config = ConfigMap(chat_template=file_spec)
        assert config.chat_template == file_spec
        assert isinstance(config.chat_template, FileSpec)

    def test_chat_template_as_file_spec_with_repo(self) -> None:
        file_spec = FileSpec("chat_template.jinja", repo="some/repo")
        config = ConfigMap(chat_template=file_spec)
        assert config.chat_template == file_spec
        assert isinstance(config.chat_template, FileSpec)
        assert config.chat_template.repo == "some/repo"

    def test_chat_template_as_json_field_spec(self) -> None:
        json_spec = JSONFieldSpec(FileSpec("config.json"), "chat_template")
        config = ConfigMap(chat_template=json_spec)
        assert config.chat_template == json_spec
        assert isinstance(config.chat_template, JSONFieldSpec)

    def test_chat_template_none(self) -> None:
        config = ConfigMap()
        assert config.chat_template is None


class TestModelSpecWithChatTemplate:
    def test_model_spec_with_string_chat_template(self) -> None:
        spec = ModelSpec(
            vendor="Test",
            family="Test",
            name="Test",
            size="1B",
            repo="test/test",
            config_type=HFLlamaConfig,
            configs=ConfigMap(chat_template=DIRECT_TEMPLATE),
        )
        assert spec.configs.chat_template == DIRECT_TEMPLATE

    def test_model_spec_with_file_spec_chat_template(self) -> None:
        spec = ModelSpec(
            vendor="Test",
            family="Test",
            name="Test",
            size="1B",
            repo="test/test",
            config_type=HFLlamaConfig,
            configs=ConfigMap(chat_template=FileSpec("chat_template.jinja")),
        )
        assert isinstance(spec.configs.chat_template, FileSpec)
        assert spec.configs.chat_template.filename == "chat_template.jinja"

    def test_model_spec_with_json_field_spec_chat_template(self) -> None:
        spec = ModelSpec(
            vendor="Test",
            family="Test",
            name="Test",
            size="1B",
            repo="test/test",
            config_type=HFLlamaConfig,
            configs=ConfigMap(chat_template=JSONFieldSpec(FileSpec("tokenizer_config.json"), "chat_template")),
        )
        assert isinstance(spec.configs.chat_template, JSONFieldSpec)
        assert spec.configs.chat_template.field_name == "chat_template"


class TestModelSpecJsonSerialization:
    def test_roundtrip_with_string_chat_template(self) -> None:
        spec = ModelSpec(
            vendor="Test",
            family="Test",
            name="Test",
            size="1B",
            repo="test/test",
            config_type=HFLlamaConfig,
            configs=ConfigMap(chat_template=DIRECT_TEMPLATE),
        )
        json_data = spec.to_json()
        restored = ModelSpec.from_json(json_data)
        assert restored.configs.chat_template == DIRECT_TEMPLATE

    def test_roundtrip_with_file_spec_chat_template(self) -> None:
        spec = ModelSpec(
            vendor="Test",
            family="Test",
            name="Test",
            size="1B",
            repo="test/test",
            config_type=HFLlamaConfig,
            configs=ConfigMap(chat_template=FileSpec("chat_template.jinja")),
        )
        json_data = spec.to_json()
        restored = ModelSpec.from_json(json_data)
        assert isinstance(restored.configs.chat_template, FileSpec)
        assert restored.configs.chat_template.filename == "chat_template.jinja"

    def test_roundtrip_with_json_field_spec_chat_template(self) -> None:
        spec = ModelSpec(
            vendor="Test",
            family="Test",
            name="Test",
            size="1B",
            repo="test/test",
            config_type=HFLlamaConfig,
            configs=ConfigMap(chat_template=JSONFieldSpec(FileSpec("config.json"), "chat_template")),
        )
        json_data = spec.to_json()
        restored = ModelSpec.from_json(json_data)
        assert isinstance(restored.configs.chat_template, JSONFieldSpec)
        assert restored.configs.chat_template.field_name == "chat_template"
        assert restored.configs.chat_template.file_spec.filename == "config.json"

    def test_from_json_with_string_chat_template(self) -> None:
        json_data = {
            "vendor": "Test",
            "family": "Test",
            "name": "Test",
            "size": "1B",
            "repo": "test/test",
            "config_type": "HFLlamaConfig",
            "configs": {
                "chat_template": DIRECT_TEMPLATE,
            },
        }
        spec = ModelSpec.from_json(json_data)
        assert spec.configs.chat_template == DIRECT_TEMPLATE

    def test_from_json_with_file_spec_chat_template(self) -> None:
        json_data = {
            "vendor": "Test",
            "family": "Test",
            "name": "Test",
            "size": "1B",
            "repo": "test/test",
            "config_type": "HFLlamaConfig",
            "configs": {
                "chat_template": {"filename": "chat_template.jinja"},
            },
        }
        spec = ModelSpec.from_json(json_data)
        assert isinstance(spec.configs.chat_template, FileSpec)
        assert spec.configs.chat_template.filename == "chat_template.jinja"

    def test_from_json_with_json_field_spec_chat_template(self) -> None:
        json_data = {
            "vendor": "Test",
            "family": "Test",
            "name": "Test",
            "size": "1B",
            "repo": "test/test",
            "config_type": "HFLlamaConfig",
            "configs": {
                "chat_template": {
                    "file_spec": {"filename": "config.json"},
                    "field_name": "chat_template",
                },
            },
        }
        spec = ModelSpec.from_json(json_data)
        assert isinstance(spec.configs.chat_template, JSONFieldSpec)
        assert spec.configs.chat_template.field_name == "chat_template"



import struct
from array import array
from collections.abc import Iterable
from dataclasses import dataclass
from itertools import chain, repeat, tee
from math import exp
from typing import Self

import xxhash

from .common import Speculator


# This is not exactly randomly distributed if size is not a power of two.
# Shouldn't matter in practice though because size <<< 2**64
def seqhash(tokens: Iterable[int], size: int) -> int:
    tokens = list(tokens)
    assert size <= 2**64
    if len(tokens) > 0:
        packed = struct.pack("<" + "I" * len(tokens), *tokens)
    else:
        packed = b""

    return xxhash.xxh3_64_intdigest(packed) % size


def padded_sliding_window(seq: Iterable[int], size: int, pad: int) -> Iterable[tuple[int, ...]]:
    seqs = tee(seq, size)
    pads = tuple(repeat(pad, size - i) for i in range(size))
    padded_seqs = tuple(chain(pad, seq) for pad, seq in zip(pads, seqs, strict=True))
    return zip(*padded_seqs, strict=False)


def softmax(logits: Iterable[float]) -> list[float]:
    logits = list(logits)
    log_max = max(logits)
    exp_logs = [exp(logit - log_max) for logit in logits]
    exp_log_sum = sum(exp_logs)
    return [exp_log / exp_log_sum for exp_log in exp_logs]


def online_mean(old_mean: float, sample: float, new_count: int) -> float:
    return old_mean + (sample - old_mean) / new_count


def update_probs(old_mean: dict[int, float], sample: dict[int, float], new_count: int, top_k: int) -> dict[int, float]:
    all_keys = set(old_mean.keys()).union(sample.keys())

    new_probs_all = {k: online_mean(old_mean.get(k, 0), sample.get(k, 0), new_count) for k in all_keys}

    new_probs_top_k = dict(sorted(new_probs_all.items(), key=lambda x: (-x[1], x[0]))[:top_k])

    new_probs_sum = sum(new_probs_top_k.values())

    new_probs_norm = {k: v / new_probs_sum for k, v in new_probs_top_k.items()}

    return new_probs_norm


@dataclass(frozen=True, eq=False)
class NGramSpeculator(Speculator):
    hashtable_size: int
    ngram_k: int
    ngram_n: int
    ngram_pad: int

    # "interior mutability" here.
    # dataclass field default_factory doesn't pass self so None here + post init as a workaround
    ngram_keys: array[int]
    ngram_values: array[float]
    ngram_counts: array[int]

    def __post_init__(self) -> None:
        if not self.hashtable_size > 0:
            raise ValueError(f"{self.hashtable_size=} (must be > 0)")
        if not self.ngram_k > 0:
            raise ValueError(f"{self.ngram_k=} (must be > 0)")
        if not self.ngram_n > 0:
            raise ValueError(f"{self.ngram_n=} (must be > 0)")

    @classmethod
    def new(cls, hashtable_size: int, ngram_k: int, ngram_n: int, ngram_pad: int = 2**32 - 1) -> Self:
        return cls(
            hashtable_size,
            ngram_k,
            ngram_n,
            ngram_pad,
            array("I", range(ngram_k)) * hashtable_size,
            array("f", repeat(0, ngram_k)) * hashtable_size,
            array("I", [0]) * hashtable_size,
        )

    def train(self, token_ids: Iterable[int], token_logits: Iterable[dict[int, float]]) -> None:
        ngram_ctx = self.ngram_n - 1
        if ngram_ctx > 0:
            contexts = padded_sliding_window(token_ids, ngram_ctx, self.ngram_pad)
        else:
            contexts = repeat(())

        for ctx, cur_logits in zip(contexts, token_logits, strict=False):
            ngram_keys, ngram_values, ngram_counts = self._seq_slice(ctx)

            ngram_counts[0] = new_count = ngram_counts[0] + 1

            old_mean = dict(zip(ngram_keys, ngram_values, strict=True))
            sample = dict(zip(cur_logits.keys(), softmax(cur_logits.values()), strict=True))

            new_probs = update_probs(old_mean, sample, new_count, self.ngram_k)

            ngram_keys[:] = array("I", new_probs.keys())
            ngram_values[:] = array("f", new_probs.values())

    def probs(self, seq: Iterable[int]) -> dict[int, float]:
        ngram_keys, ngram_values, _ = self._seq_slice(seq)
        return dict(zip(ngram_keys, ngram_values, strict=True))

    # python < 3.13 doesn't support memoryview[T], but if T is not specified typechecker incorrectly assumes it's int
    def _seq_slice(self, seq: Iterable[int]) -> tuple["memoryview[int]", "memoryview[float]", "memoryview[int]"]:
        seq = list(seq)
        ngram_ctx = self.ngram_n - 1
        if ngram_ctx > 0:
            padded_seq = [*repeat(self.ngram_pad, max(ngram_ctx - len(seq), 0)), *seq[-ngram_ctx:]]
        else:
            padded_seq = []

        seq_hash = seqhash(padded_seq, self.hashtable_size)
        idx_start = seq_hash * self.ngram_k
        idx_end = seq_hash * self.ngram_k + self.ngram_k

        return (
            memoryview(self.ngram_keys)[idx_start:idx_end],
            memoryview(self.ngram_values)[idx_start:idx_end],  # type: ignore (typechecker bug)
            memoryview(self.ngram_counts)[seq_hash : (seq_hash + 1)],
        )

    def serialize(self) -> bytes:
        hdr = struct.pack("<4I", self.hashtable_size, self.ngram_k, self.ngram_n, self.ngram_pad)
        return hdr + bytes(self.ngram_keys) + bytes(self.ngram_values) + bytes(self.ngram_counts)

    @classmethod
    def deserialize(cls, blob: bytes) -> Self:
        offset = 16

        hashtable_size, ngram_k, ngram_len, ngram_pad = struct.unpack("<4I", blob[:offset])

        ngram_kv_len = 4 * ngram_k * hashtable_size

        ngram_keys = array("I", blob[offset : offset + ngram_kv_len])
        offset += ngram_kv_len

        ngram_values = array("f", blob[offset : offset + ngram_kv_len])
        offset += ngram_kv_len

        ngram_counts_len = 4 * hashtable_size
        ngram_counts = array("I", blob[offset : offset + ngram_counts_len])
        offset += ngram_counts_len

        return cls(hashtable_size, ngram_k, ngram_len, ngram_pad, ngram_keys, ngram_values, ngram_counts)



import json
import pathlib
import shutil
import tempfile
from collections.abc import Generator
from contextlib import contextmanager
from pathlib import Path
from typing import Any

import pytest
import torch
from safetensors.flax import save_file

from lalamo.common import flatten_parameters
from lalamo.model_import import REPO_TO_MODEL, ModelMetadata, import_model
from lalamo.model_import.model_specs import ModelType
from lalamo.model_import.model_specs.lfm2 import LFM2_MODELS
from lalamo.models import ClassifierModelConfig, LanguageModelConfig
from lalamo.modules import config_converter
from tests.test_models import DType, ModelTestSpec

MODEL_LIST: list[ModelTestSpec] = [
    ModelTestSpec("trymirai/chat-moderation-router", DType.FLOAT32),
    ModelTestSpec("Qwen/Qwen3-0.6B", DType.FLOAT32),
    ModelTestSpec("Qwen/Qwen2.5-0.5B-Instruct", DType.FLOAT32),
    ModelTestSpec("google/gemma-3-1b-it", DType.FLOAT32),
    ModelTestSpec("deepseek-ai/DeepSeek-R1-Distill-Qwen-1.5B", DType.FLOAT32),
    ModelTestSpec("meta-llama/Llama-3.2-1B-Instruct", DType.FLOAT32),
    ModelTestSpec("cartesia-ai/Llamba-1B", DType.FLOAT32),
    ModelTestSpec("cartesia-ai/Llamba-1B-4bit-mlx", DType.FLOAT32),
] + \
[ModelTestSpec(model.repo, DType.FLOAT32) for model in LFM2_MODELS]

MODEL_LIST += (
    [
        ModelTestSpec("Qwen/Qwen3-4B-AWQ", DType.FLOAT16),
        ModelTestSpec("openai/gpt-oss-20b", DType.FLOAT16),
    ]
    if torch.cuda.is_available()
    else []
)


@contextmanager
def temporary_directory() -> Generator[Path, Any, None]:
    """
    Context manager that creates a temporary directory and ensures cleanup.

    Yields:
        Path: Path to the temporary directory
    """
    temp_dir = Path(tempfile.mkdtemp())

    try:
        yield temp_dir
    finally:
        # Clean up temporary directory
        if temp_dir.exists():
            shutil.rmtree(temp_dir)


@pytest.mark.parametrize("test_spec", MODEL_LIST, ids=[m.model_repo for m in MODEL_LIST])
def test_model_conversion(test_spec: ModelTestSpec, tmp_path: pathlib.Path) -> None:
    """
    Test the conversion -> export -> import flow for a given model.
    NOTE: Using tmp_path fixture provided by pytest runtime environment.
    """

    model_repo_name = test_spec.model_repo
    model_dtype = test_spec.dtype
    assert model_dtype is not None

    model_repo = REPO_TO_MODEL.get(model_repo_name)
    assert model_repo, f"Unknown model specified: {model_repo_name}"

    # Step 1: Import model from HF format
    model, metadata = import_model(
        model_spec=model_repo,
        precision=model_dtype.jax_dtype,
    )

    # Step 2: Export model
    model.message_processor.tokenizer.save(str(tmp_path / "tokenizer.json"))
    weights = flatten_parameters(model.export_weights())
    del model
    save_file(weights, tmp_path / "model.safetensors")
    config_json = config_converter.unstructure(metadata, ModelMetadata)
    with open(tmp_path / "config.json", "w") as file:
        json.dump(config_json, file, indent=4)

    # Step 3: Re-import model from Lalamo format
    model = None
    match metadata.model_type:
        case ModelType.LANGUAGE_MODEL:
            model = LanguageModelConfig.load_model(tmp_path)
        case ModelType.CLASSIFIER_MODEL:
            model = ClassifierModelConfig.load_model(tmp_path)
    assert model is not None, f"Failed to load model {model_repo_name}"
    del model



from textwrap import dedent

import jax.numpy as jnp
import pytest
from attr import dataclass
from jaxtyping import Array, Int
from transformers import AutoTokenizer
from transformers.tokenization_utils import PreTrainedTokenizer

from lalamo.model_import import REPO_TO_MODEL, import_model
from lalamo.models import LanguageModel
from lalamo.sampling import GreedyPolicy


@dataclass(frozen=True)
class GenerationInput:
    prompt: str
    token_ids: Int[Array, " tokens"]


@pytest.fixture
def generation_input(tokenizer: PreTrainedTokenizer) -> GenerationInput:
    prompt = dedent(
        """
        <|im_start|>user
        hello<|im_end|>
        <|im_start|>assistant
        <think>\n\n</think>\n\n
    """.lstrip(),
    )
    tokens = jnp.array(tokenizer.encode(prompt))
    return GenerationInput(prompt, tokens)


@pytest.fixture
def another_generation_input(tokenizer: PreTrainedTokenizer) -> GenerationInput:
    prompt = dedent(
        """
        <|im_start|>user
        How are you?<|im_end|>
        <|im_start|>assistant
        <think>\n\n</think>\n\n
    """.lstrip(),
    )
    tokens = jnp.array(tokenizer.encode(prompt))
    return GenerationInput(prompt, tokens)


@pytest.fixture
def language_model() -> LanguageModel:
    model = import_model(REPO_TO_MODEL["Qwen/Qwen3-0.6B"]).model
    assert isinstance(model, LanguageModel)
    return model


@pytest.fixture
def tokenizer() -> PreTrainedTokenizer:
    return AutoTokenizer.from_pretrained("Qwen/Qwen3-0.6B")


def test_tokenizer(language_model: LanguageModel, generation_input: GenerationInput) -> None:
    token_ids = language_model.message_processor.tokenize_text(generation_input.prompt)
    ref_token_ids = generation_input.token_ids.tolist()
    assert token_ids == ref_token_ids


@pytest.mark.parametrize("num_top_logits_to_return", [None, 8, 16])
def test_eager_generation(
    language_model: LanguageModel,
    tokenizer: PreTrainedTokenizer,
    generation_input: GenerationInput,
    num_top_logits_to_return: int | None,
) -> None:
    result = language_model.generate_tokens(
        generation_input.token_ids[None, :],
        max_output_length=32,
        num_top_logits_to_return=num_top_logits_to_return,
    )
    token_ids = result.token_ids.squeeze(0)
    response_text = tokenizer.decode(token_ids)  # type: ignore
    assert "<|im_end|>" in response_text

    if num_top_logits_to_return is not None:
        assert result.top_k_token_ids is not None
        assert result.top_k_token_logits is not None

        expected_shape = (1, result.token_ids.shape[1], num_top_logits_to_return)
        assert result.top_k_token_ids.shape == expected_shape
        assert result.top_k_token_logits.shape == expected_shape

        top_k_token_ids = result.top_k_token_ids.squeeze(0).tolist()
        top_k_token_logits = result.top_k_token_logits.squeeze(0).tolist()

        eos_id = tokenizer.encode("<|im_end|>")[0]
        eos_idx = token_ids.tolist().index(eos_id)
        assert top_k_token_ids[eos_idx][0] == eos_id
        assert top_k_token_logits[eos_idx][0] > max(top_k_token_logits[eos_idx][1:])
    else:
        assert result.top_k_token_ids is None
        assert result.top_k_token_logits is None


def test_padding(language_model: LanguageModel, tokenizer: PreTrainedTokenizer) -> None:
    prompt = dedent(
        """
        <|im_start|>user
        Talk about elephants<|im_end|>
        <|im_start|>assistant
        <think>\n\n</think>\n\n
    """.lstrip(),
    )
    token_ids = jnp.array(tokenizer.encode(prompt))[None, :]

    response_token_ids = language_model.generate_tokens(
        token_ids,
        prompt_lengths_without_padding=jnp.array([0], dtype=jnp.int32),
        max_output_length=32,
    ).token_ids.squeeze(0)
    response_text = tokenizer.decode(response_token_ids)  # type: ignore
    assert "elephants" not in response_text.lower()

    response_token_ids = language_model.generate_tokens(
        token_ids,
        prompt_lengths_without_padding=jnp.array([token_ids.size]),
        max_output_length=32,
    ).token_ids.squeeze(0)
    response_text = tokenizer.decode(response_token_ids)  # type: ignore
    assert "elephants" in response_text.lower()


def test_batch_generation(
    language_model: LanguageModel,
    tokenizer: PreTrainedTokenizer,
    generation_input: GenerationInput,
    another_generation_input: GenerationInput,
) -> None:
    inputs = [generation_input, another_generation_input]
    pad_token_id = 0

    max_len = max(inp.token_ids.size for inp in inputs)

    batched_prompt_lengths = jnp.array([inp.token_ids.size for inp in inputs])
    padded_token_ids = jnp.array(
        [
            jnp.pad(
                inp.token_ids,
                (0, max_len - inp.token_ids.size),
                constant_values=pad_token_id,
            )
            for inp in inputs
        ],
    )

    response_token_ids = language_model.generate_tokens(
        padded_token_ids,
        prompt_lengths_without_padding=batched_prompt_lengths,
        max_output_length=32,
    ).token_ids
    for ids in response_token_ids:
        response_text = tokenizer.decode(ids)
        assert "<|im_end|>" in response_text


def test_streaming_generation(
    language_model: LanguageModel,
    tokenizer: PreTrainedTokenizer,
    generation_input: GenerationInput,
) -> None:
    token_stream = language_model.stream_tokens(generation_input.token_ids, max_output_length=32)
    response_token_ids = jnp.array(list(token_stream))
    response_text = tokenizer.decode(response_token_ids)  # type: ignore
    assert "<|im_end|>" in response_text


def test_streaming_vs_eager_consistency(
    language_model: LanguageModel,
    generation_input: GenerationInput,
) -> None:
    sampling_policy = GreedyPolicy()
    eager_token_ids = language_model.generate_tokens(
        generation_input.token_ids[None, :],
        sampling_policy=sampling_policy,
        max_output_length=32,
        eos_token_ids=jnp.array([-1]),  # Never stop.
    ).token_ids.squeeze(0)

    streaming_token_generator = language_model.stream_tokens(
        generation_input.token_ids,
        sampling_policy=sampling_policy,
        max_output_length=32,
        eos_token_ids=jnp.array([-1]),  # Never stop.
    )
    streaming_token_ids = jnp.array(list(streaming_token_generator))

    assert jnp.array_equal(eager_token_ids, streaming_token_ids), (
        eager_token_ids.squeeze().tolist(),
        streaming_token_ids.squeeze().tolist(),
    )



from lalamo.main import collect_traces, convert, estimate_batchsize, train
from lalamo.message_processor import (
    AssistantMessage,
    ContentBlock,
    Image,
    Message,
    SystemMessage,
    ToolSchema,
    UserMessage,
)
from lalamo.model_import import ModelSpec, import_model
from lalamo.models import ClassifierModel, LanguageModel
from lalamo.speculator import (
    CollectTracesEvent,
    SpeculatorTrainingEvent,
)

__version__ = "0.5.11"

__all__ = [
    "AssistantMessage",
    "ClassifierModel",
    "CollectTracesEvent",
    "ContentBlock",
    "Image",
    "LanguageModel",
    "Message",
    "ModelSpec",
    "SpeculatorTrainingEvent",
    "SystemMessage",
    "ToolSchema",
    "UserMessage",
    "collect_traces",
    "convert",
    "estimate_batchsize",
    "import_model",
    "train",
]



import jax.numpy as jnp
import torch
from jaxtyping import Array

__all__ = ["jax_to_torch", "torch_to_jax"]


@torch.no_grad()
def _torch_to_jax_bfloat16(tensor: torch.Tensor) -> Array:
    if tensor.dtype != torch.bfloat16:
        raise ValueError("Trying to convert non-bfloat16 tensor to bfloat16")
    intermediate_tensor = tensor.view(torch.uint16)
    return jnp.array(intermediate_tensor).view("bfloat16")


def torch_to_jax(array: torch.Tensor) -> Array:
    array = array.detach().cpu()
    if array.dtype == torch.bfloat16:
        return _torch_to_jax_bfloat16(array)
    return jnp.array(array.numpy())


def jax_to_torch(array: Array) -> torch.Tensor:
    from torch.utils import dlpack as _dlpack

    if array.dtype == jnp.bfloat16:
        intermediate_array = array.view(jnp.uint16)
        return _dlpack.from_dlpack(intermediate_array).view(torch.bfloat16)
    return _dlpack.from_dlpack(array)



import importlib.metadata
import json
from collections import ChainMap
from collections.abc import Callable
from contextlib import ExitStack
from dataclasses import dataclass
from pathlib import Path
from typing import NamedTuple

import huggingface_hub
import jax.numpy as jnp
from jax import Array
from jaxtyping import DTypeLike
from tokenizers import Tokenizer

from lalamo.message_processor import MessageProcessor, MessageProcessorConfig
from lalamo.models import ClassifierModel, ClassifierModelConfig, GenerationConfig, LanguageModel, LanguageModelConfig
from lalamo.modules import Classifier, Decoder, LalamoModule
from lalamo.quantization import QuantizationMode
from lalamo.utils import process_chat_template

from .decoder_configs import ForeignClassifierConfig, ForeignConfig, ForeignLMConfig
from .huggingface_generation_config import HFGenerationConfig
from .huggingface_tokenizer_config import HFTokenizerConfig
from .model_specs import REPO_TO_MODEL, FileSpec, ModelSpec, ModelType, UseCase
from .model_specs.common import JSONFieldSpec

__all__ = [
    "REPO_TO_MODEL",
    "DownloadingFileEvent",
    "FinishedDownloadingFileEvent",
    "InitializingModelEvent",
    "ModelMetadata",
    "ModelSpec",
    "ModelType",
    "StatusEvent",
    "import_model",
]


LALAMO_VERSION = importlib.metadata.version("lalamo")


class DownloadingFileEvent(NamedTuple):
    file: FileSpec


class FinishedDownloadingFileEvent(NamedTuple):
    file: FileSpec


class InitializingModelEvent(NamedTuple):
    pass


class FinishedInitializingModelEvent(NamedTuple):
    pass


type StatusEvent = (
    DownloadingFileEvent | FinishedDownloadingFileEvent | InitializingModelEvent | FinishedInitializingModelEvent
)


@dataclass(frozen=True)
class ModelMetadata:
    toolchain_version: str
    vendor: str
    family: str
    name: str
    size: str
    quantization: QuantizationMode | None
    repo: str
    use_cases: tuple[UseCase, ...]
    model_type: ModelType
    model_config: LanguageModelConfig | ClassifierModelConfig
    grammar_start_tokens: tuple[str, ...]


def download_file(
    file_spec: FileSpec,
    model_repo: str,
    output_dir: Path | str | None = None,
    progress_callback: Callable[[StatusEvent], None] | None = None,
) -> Path:
    if progress_callback is not None:
        progress_callback(DownloadingFileEvent(file_spec))
    result = huggingface_hub.hf_hub_download(
        repo_id=file_spec.repo or model_repo,
        local_dir=output_dir,
        filename=file_spec.filename,
    )
    if progress_callback is not None:
        progress_callback(FinishedDownloadingFileEvent(file_spec))
    return Path(result)


def list_weight_files(model_repo: str) -> list[FileSpec]:
    all_files = huggingface_hub.list_repo_files(model_repo)
    return [FileSpec(filename) for filename in all_files if filename.endswith(".safetensors")]


def download_weights(
    model_spec: ModelSpec,
    output_dir: Path | str | None = None,
    progress_callback: Callable[[StatusEvent], None] | None = None,
) -> list[Path]:
    return [
        download_file(file_spec, model_spec.repo, output_dir, progress_callback)
        for file_spec in list_weight_files(model_spec.repo)
    ]


def download_config_file(
    model_spec: ModelSpec,
    output_dir: Path | str | None = None,
    progress_callback: Callable[[StatusEvent], None] | None = None,
) -> Path:
    return download_file(model_spec.configs.model_config, model_spec.repo, output_dir, progress_callback)


class ImportResults(NamedTuple):
    model: LanguageModel | ClassifierModel
    metadata: ModelMetadata


def import_message_processor(
    model_spec: ModelSpec,
    output_dir: Path | str | None = None,
    progress_callback: Callable[[StatusEvent], None] | None = None,
) -> MessageProcessor:
    tokenizer_file = download_file(model_spec.configs.tokenizer, model_spec.repo, output_dir, progress_callback)
    tokenizer_config_file = download_file(
        model_spec.configs.tokenizer_config,
        model_spec.repo,
        output_dir,
        progress_callback,
    )
    tokenizer_config = HFTokenizerConfig.from_json(tokenizer_config_file)
    if tokenizer_config.chat_template is None:
        match model_spec.configs.chat_template:
            case JSONFieldSpec(file_spec, field_name):
                json_file = download_file(file_spec, model_spec.repo, output_dir)
                with open(json_file) as file:
                    json_dict = json.load(file)
                prompt_template = json_dict[field_name]
            case FileSpec(_) as file_spec:
                chat_template_file = download_file(file_spec, model_spec.repo, output_dir)
                prompt_template = chat_template_file.read_text()
            case str() as template_string:
                prompt_template = template_string
            case None:
                raise ValueError("No chat template specified.")
    else:
        if model_spec.configs.chat_template is not None:
            raise ValueError("Conflicting chat template specifications.")
        prompt_template = tokenizer_config.chat_template
    prompt_template = process_chat_template(prompt_template)
    tokenizer = Tokenizer.from_file(str(tokenizer_file))

    added_tokens = tokenizer_config.added_tokens()
    added_special_tokens = [token for token in added_tokens if token.special]
    added_not_special_tokens = [token for token in added_tokens if not token.special]
    tokenizer.add_special_tokens(added_special_tokens)
    tokenizer.add_tokens(added_not_special_tokens)

    message_processor_config = MessageProcessorConfig(
        prompt_template=prompt_template,
        output_parser_regex=model_spec.output_parser_regex,
        system_role_name=model_spec.system_role_name,
        user_role_name=model_spec.user_role_name,
        assistant_role_name=model_spec.assistant_role_name,
        bos_token=tokenizer_config.bos_token,
    )
    return MessageProcessor(config=message_processor_config, tokenizer=tokenizer)


def _load_main_processing_module(
    model_spec: ModelSpec,
    precision: DTypeLike,
    foreign_config: ForeignConfig,
    progress_callback: Callable[[StatusEvent], None] | None = None,
    context_length: int | None = None,
    accumulation_precision: DTypeLike = jnp.float32,
) -> LalamoModule:
    weights_paths = download_weights(model_spec, progress_callback=progress_callback)
    with ExitStack() as stack:
        weights_shards = []
        metadata_shards = []
        for weights_path in weights_paths:
            weights_shard, metadata_shard = stack.enter_context(model_spec.weights_type.load(weights_path, precision))
            weights_shards.append(weights_shard)
            metadata_shards.append(metadata_shard)
        weights_dict: ChainMap[str, Array] = ChainMap(*weights_shards)
        metadata_dict: ChainMap[str, str] = ChainMap(*metadata_shards)

        if progress_callback is not None:
            progress_callback(InitializingModelEvent())

        processing_module = foreign_config.load(
            context_length,
            precision,
            accumulation_precision,
            weights_dict,
            metadata_dict,
        )

    return processing_module


def _import_language_model(
    model_spec: ModelSpec,
    *,
    context_length: int | None = None,
    precision: DTypeLike | None = None,
    accumulation_precision: DTypeLike = jnp.float32,
    progress_callback: Callable[[StatusEvent], None] | None = None,
) -> tuple[LanguageModel, LanguageModelConfig]:
    foreign_decoder_config_file = download_config_file(model_spec)
    foreign_decoder_config = model_spec.config_type.from_json(foreign_decoder_config_file)
    assert isinstance(foreign_decoder_config, ForeignLMConfig)

    if precision is None:
        precision = foreign_decoder_config.default_precision
    decoder = _load_main_processing_module(
        model_spec,
        precision,
        foreign_decoder_config,
        progress_callback,
        context_length,
        accumulation_precision,
    )
    assert isinstance(decoder, Decoder)

    if progress_callback is not None:
        progress_callback(FinishedInitializingModelEvent())

    message_processor = import_message_processor(model_spec)

    stop_token_ids = tuple(foreign_decoder_config.eos_token_ids)

    if model_spec.configs.generation_config is not None:
        hf_generation_config_file = download_file(model_spec.configs.generation_config, model_spec.repo)
        hf_generation_config = HFGenerationConfig.from_json(hf_generation_config_file)
        generation_config = GenerationConfig(
            stop_token_ids=stop_token_ids,
            temperature=hf_generation_config.temperature,
            top_p=hf_generation_config.top_p,
            top_k=hf_generation_config.top_k,
            banned_tokens=None,
        )
    else:
        generation_config = GenerationConfig(
            stop_token_ids=stop_token_ids,
            temperature=None,
            top_p=None,
            top_k=None,
            banned_tokens=None,
        )

    language_model_config = LanguageModelConfig(
        model_config=decoder.config,
        message_processor_config=message_processor.config,
        generation_config=generation_config,
    )

    language_model = LanguageModel(language_model_config, decoder, message_processor)
    return language_model, language_model_config


def _import_classifier(
    model_spec: ModelSpec,
    *,
    context_length: int | None = None,
    precision: DTypeLike | None = None,
    accumulation_precision: DTypeLike = jnp.float32,
    progress_callback: Callable[[StatusEvent], None] | None = None,
) -> tuple[ClassifierModel, ClassifierModelConfig]:
    foreign_classifier_config_file = download_config_file(model_spec)
    foreign_classifier_config = model_spec.config_type.from_json(foreign_classifier_config_file)
    assert isinstance(foreign_classifier_config, ForeignClassifierConfig)

    if precision is None:
        precision = foreign_classifier_config.default_precision

    classifier = _load_main_processing_module(
        model_spec,
        precision,
        foreign_classifier_config,
        progress_callback,
        context_length,
        accumulation_precision,
    )
    assert isinstance(classifier, Classifier)

    if progress_callback is not None:
        progress_callback(FinishedInitializingModelEvent())

    message_processor = import_message_processor(model_spec)

    classifier_model_config = ClassifierModelConfig(
        model_config=classifier.config,
        message_processor_config=message_processor.config,
    )
    classifier_model = ClassifierModel(classifier_model_config, classifier, message_processor)
    return classifier_model, classifier_model_config


def import_model(
    model_spec: ModelSpec | str,
    *,
    context_length: int | None = None,
    precision: DTypeLike | None = None,
    accumulation_precision: DTypeLike = jnp.float32,
    progress_callback: Callable[[StatusEvent], None] | None = None,
) -> ImportResults:
    if isinstance(model_spec, str):
        try:
            model_spec = REPO_TO_MODEL[model_spec]
        except KeyError as e:
            raise ValueError(f"Unknown model: {model_spec}") from e

    match model_spec.model_type:
        case ModelType.LANGUAGE_MODEL:
            model, config = _import_language_model(
                model_spec,
                context_length=context_length,
                precision=precision,
                accumulation_precision=accumulation_precision,
                progress_callback=progress_callback,
            )
        case ModelType.CLASSIFIER_MODEL:
            model, config = _import_classifier(
                model_spec,
                context_length=context_length,
                precision=precision,
                accumulation_precision=accumulation_precision,
                progress_callback=progress_callback,
            )

    metadata = ModelMetadata(
        toolchain_version=LALAMO_VERSION,
        vendor=model_spec.vendor,
        family=model_spec.family,
        name=model_spec.name,
        size=model_spec.size,
        quantization=model_spec.quantization,
        repo=model_spec.repo,
        use_cases=model_spec.use_cases,
        model_type=model_spec.model_type,
        model_config=config,
        grammar_start_tokens=model_spec.grammar_start_tokens,
    )
    return ImportResults(model, metadata)



import unittest

import pytest

from tests.test_models import MLX_AVAILABLE, DType, ModelTestSpec, _test_model

if MLX_AVAILABLE:
    from tests.mlx_tracer import MLXDecoderTracer

# token_stride=1 is required because mlx doesn't accept token positions
MODEL_LIST = [
    ModelTestSpec("Qwen/Qwen3-0.6B-MLX-4bit", DType.FLOAT32, token_stride=1),
    ModelTestSpec("Qwen/Qwen3-4B-MLX-4bit", DType.FLOAT32, token_stride=1),
]


@unittest.skipUnless(MLX_AVAILABLE, "requires mlx")
@pytest.mark.parametrize("test_spec", MODEL_LIST, ids=[m.model_repo for m in MODEL_LIST])
def test_mlx_model(test_spec: ModelTestSpec) -> None:
    _test_model(test_spec, MLXDecoderTracer)  # type: ignore



import importlib.util
import unittest

import pytest

from tests.test_models import DType, ModelTestSpec, _test_model

CMX_AVAILABLE = importlib.util.find_spec("cartesia_mlx")

if CMX_AVAILABLE:
    from tests.cartesia_mlx_tracer import CartesiaMLXDecoderTracer

MODEL_LIST = [
    ModelTestSpec("cartesia-ai/Llamba-1B-4bit-mlx", DType.FLOAT32, token_stride=1),
]


@unittest.skipUnless(CMX_AVAILABLE, "requires mlx")
@pytest.mark.parametrize("test_spec", MODEL_LIST, ids=[m.model_repo for m in MODEL_LIST])
def test_cartesia_mlx_model(test_spec: ModelTestSpec) -> None:
    _test_model(test_spec, CartesiaMLXDecoderTracer) # type: ignore




# from .executorch import load_executorch
from .huggingface import load_huggingface_classifier, load_huggingface_decoder

__all__ = [
    "load_huggingface_classifier",
    # "load_executorch",
    "load_huggingface_decoder",
]



from abc import abstractmethod
from dataclasses import dataclass
from enum import Enum
from types import UnionType
from typing import Any, Generic, Self, TypeVar

import equinox as eqx
from cattrs import Converter
from jax import numpy as jnp
from jaxtyping import Array, DTypeLike

from lalamo.common import ParameterTree

__all__ = [
    "DummyUnionMember",
    "ForwardPassMode",
    "LalamoModule",
    "PositionalEmbeddingSelector",
    "config_converter",
    "register_config_union",
]


class PositionalEmbeddingSelector(Enum):
    GLOBAL = "global"
    LOCAL = "sliding_window"
    NONE = "none"


class ForwardPassMode(Enum):
    MULTI_TOKEN = "multi_token"
    SINGLE_TOKEN = "single_token"


ConfigT_co = TypeVar("ConfigT_co", covariant=True)


class LalamoModule(eqx.Module, Generic[ConfigT_co]):  # noqa: UP046
    config: ConfigT_co = eqx.field(static=True)

    @property
    @abstractmethod
    def activation_precision(self) -> DTypeLike: ...

    @abstractmethod
    def export_weights(self) -> ParameterTree[Array]: ...

    @abstractmethod
    def import_weights(
        self,
        weights: ParameterTree[Array],
    ) -> Self: ...


def _dtype_to_str(dtype: DTypeLike) -> str:
    if dtype == jnp.bfloat16:
        return "bfloat16"
    try:
        return str(dtype.dtype)  # type: ignore
    except AttributeError:
        return str(dtype)


def _str_to_dtype(dtype_str: str) -> jnp.dtype:
    return {
        "int4": jnp.int4,
        "int8": jnp.int8,
        "int16": jnp.int16,
        "int32": jnp.int32,
        "int64": jnp.int64,
        "bfloat16": jnp.bfloat16,
        "float16": jnp.float16,
        "float32": jnp.float32,
        "float64": jnp.float64,
    }[dtype_str]


config_converter = Converter()


config_converter.register_unstructure_hook_func(
    lambda t: t in [jnp.dtype, DTypeLike],
    _dtype_to_str,
)

config_converter.register_structure_hook_func(
    lambda t: t in [jnp.dtype, DTypeLike],
    lambda s, _: _str_to_dtype(s),
)


def register_config_union(union_type: UnionType) -> None:
    union_members = union_type.__args__
    name_to_type = {m.__name__: m for m in union_members}

    def unstructure(obj: object) -> dict | None:
        if obj is None:
            return None
        return {
            "type": obj.__class__.__name__,
            **config_converter.unstructure(obj),
        }

    config_converter.register_unstructure_hook(
        union_type,
        unstructure,
    )

    config_converter.register_unstructure_hook(
        union_type | None,
        unstructure,
    )

    def structure[T](config: dict | None, _: type[T]) -> T | None:
        if config is None:
            return None
        new_config = dict(config)
        type_name = new_config.pop("type")
        target_type = name_to_type[type_name]
        return config_converter.structure(new_config, target_type)

    config_converter.register_structure_hook(
        union_type,
        structure,
    )

    config_converter.register_structure_hook(
        union_type | None,
        structure,
    )


@dataclass
class DummyUnionMember:
    def __getattribute__(self, name: str, /) -> Any:  # noqa: ANN401
        raise NotImplementedError

    def __call__(self, *args: Any, **kwargs: Any) -> Any:  # noqa: ANN401
        raise NotImplementedError



from abc import abstractmethod
from typing import Self

import equinox as eqx
from jax.tree_util import register_pytree_node_class

from lalamo.common import ParameterTree

__all__ = ["State", "StateLayerBase"]


class StateLayerBase(eqx.Module):
    @abstractmethod
    def export(self) -> ParameterTree: ...


@register_pytree_node_class
class State(tuple[StateLayerBase, ...]):
    __slots__ = ()

    def tree_flatten(self) -> tuple[tuple[StateLayerBase, ...], None]:
        return (tuple(self), None)

    @classmethod
    def tree_unflatten(cls, aux_data: None, children: tuple[StateLayerBase, ...]) -> Self:  # noqa: ARG003
        return cls(children)



import random
from collections.abc import Callable, Iterable
from typing import NamedTuple

from lalamo.data.lalamo_completions import LalamoCompletion
from lalamo.speculator.common import Speculator


class SpeculatorTrainingEvent(NamedTuple):
    trained_sequences: int
    trained_tokens: int


def train_speculator(
    speculator: Speculator,
    traces: Iterable[LalamoCompletion],
    tokens_to_train: int | None = None,
    progress_callback: Callable[[SpeculatorTrainingEvent], None] | None = None,
) -> None:
    trained_tokens = 0

    for trained_sequences, trace in enumerate(traces, start=1):
        if tokens_to_train is not None and trained_tokens + len(trace.completion_token_ids) > tokens_to_train:
            end = tokens_to_train - trained_tokens
        else:
            end = None
        token_ids = trace.completion_token_ids[:end]
        token_logits = trace.completion_token_logits[:end]

        speculator.train(token_ids, token_logits)

        trained_tokens += len(token_ids)

        if progress_callback is not None:
            progress_callback(SpeculatorTrainingEvent(trained_sequences, trained_tokens))

        if tokens_to_train is not None and trained_tokens >= tokens_to_train:
            break


def test_speculator(
    speculator: Speculator,
    sequence: Iterable[int] = [],
    max_completion_length: int = 32,
) -> list[int]:
    sequence = list(sequence)
    for _ in range(max_completion_length):
        probs = speculator.probs(sequence)
        if sum(probs.values()) == 0:
            break

        selected = random.choices(list(probs.keys()), weights=list(probs.values()), k=1)[0]
        sequence.append(selected)
    return sequence



from lalamo.model_import.decoder_configs import HFLlamaConfig

from .common import ModelSpec

__all__ = ["HUGGINGFACE_MODELS"]

HUGGINGFACE_MODELS = [
    ModelSpec(
        vendor="HuggingFace",
        family="SmolLM2",
        name="SmolLM2-1.7B-Instruct",
        size="1.7B",
        quantization=None,
        repo="HuggingFaceTB/SmolLM2-1.7B-Instruct",
        config_type=HFLlamaConfig,
        use_cases=tuple(),
    ),
]



from collections.abc import Mapping
from dataclasses import dataclass, field
from typing import Literal

import jax.numpy as jnp
from jaxtyping import DTypeLike

from lalamo.modules import DecoderConfig, TiedEmbeddingConfig, TransformerConfig
from lalamo.modules.activations import GELU
from lalamo.modules.linear import FullPrecisionLinearConfig
from lalamo.modules.mlp import DenseMLPConfig
from lalamo.modules.normalization import NormalizationConfig, UpcastMode
from lalamo.modules.rope import LinearScalingRoPEConfig, UnscaledRoPEConfig, YARNRoPEConfig
from lalamo.modules.token_mixers.attention import AttentionConfig
from lalamo.modules.transformer_layer import TransformerLayerConfig

from .common import HuggingFaceLMConfig

__all__ = ["HFGemma3Config", "HFGemma3TextConfig"]


def _round_to_bfloat16(x: float) -> float:
    return jnp.asarray(x).astype(jnp.bfloat16).item()


@dataclass(frozen=True)
class GemmaRoPEScalingConfig:
    factor: float
    rope_type: Literal["linear"]


@dataclass(frozen=True)
class YarnRopeScalingConfig:
    factor: float
    beta_fast: float
    beta_slow: float
    original_max_position_embeddings: int
    rope_type: Literal["yarn"]
    truncate: bool = False


@dataclass(frozen=True)
class HFGemma3TextConfigRaw:
    hidden_size: int
    intermediate_size: int
    model_type: Literal["gemma3_text"]
    num_hidden_layers: int
    sliding_window: int
    sliding_window_pattern: int
    rms_norm_eps: float = 1e-06
    query_pre_attn_scalar: float = 256.0
    attention_bias: bool = False
    num_attention_heads: int = 8
    num_key_value_heads: int = 4
    attn_logit_softcapping: float | None = None
    head_dim: int = 256
    max_position_embeddings: int = 131072
    rope_theta: float = 1000000.0
    rope_local_base_freq: float = 10000.0
    rope_scaling: GemmaRoPEScalingConfig | YarnRopeScalingConfig | None = None
    final_logit_softcapping: float | None = None
    vocab_size: int = 262208

    @property
    def sliding_window_sizes(self) -> list[int | None]:
        result = []
        for i in range(self.num_hidden_layers):
            if (i + 1) % self.sliding_window_pattern == 0:
                result.append(None)
            else:
                result.append(self.sliding_window)
        return result

    def to_decoder_config(
        self,
        context_length: int | None,
        activation_precision: DTypeLike,
        accumulation_precision: DTypeLike,
        metadata_dict: Mapping[str, str],  # noqa: ARG002
    ) -> DecoderConfig:
        input_scale = _round_to_bfloat16(self.hidden_size**0.5)
        attention_scale = self.query_pre_attn_scalar**-0.5
        embedding_config = TiedEmbeddingConfig(
            input_scale=input_scale,
            logit_soft_cap=self.final_logit_softcapping,
            precision=activation_precision,
        )
        rms_norm_config = NormalizationConfig(
            scale_precision=activation_precision,
            accumulation_precision=accumulation_precision,
            epsilon=self.rms_norm_eps,
            scale_offset=1.0,
            upcast_mode=UpcastMode.FULL_LAYER,
            subtract_mean=False,
        )

        if isinstance(self.rope_scaling, GemmaRoPEScalingConfig):
            global_rope_config = LinearScalingRoPEConfig(
                precision=activation_precision,
                base=self.rope_theta,
                max_sequence_length=self.max_position_embeddings,
                scaling_factor=self.rope_scaling.factor,
            )
        elif isinstance(self.rope_scaling, YarnRopeScalingConfig):
            global_rope_config = YARNRoPEConfig(
                precision=activation_precision,
                base=self.rope_theta,
                scaling_factor=self.rope_scaling.factor,
                max_sequence_length=self.max_position_embeddings,
                original_context_length=self.rope_scaling.original_max_position_embeddings,
                beta_fast=self.rope_scaling.beta_fast,
                beta_slow=self.rope_scaling.beta_slow,
                truncate=self.rope_scaling.truncate,
            )
        elif self.rope_scaling is None:
            global_rope_config = UnscaledRoPEConfig(
                precision=activation_precision,
                base=self.rope_theta,
                max_sequence_length=context_length or self.max_position_embeddings,
            )
        else:
            raise ValueError("Invalid rope scaling configuration")

        local_rope_config = UnscaledRoPEConfig(
            precision=activation_precision,
            base=self.rope_local_base_freq,
            max_sequence_length=context_length or self.max_position_embeddings,
        )

        linear_config = FullPrecisionLinearConfig(precision=activation_precision)
        mlp_config = DenseMLPConfig(
            linear_config=linear_config,
            activation=GELU(),
            has_up_biases=False,
            has_down_biases=False,
            up_clipping=None,
            gate_clipping=None,
        )
        layer_configs = []
        for sliding_window_size in self.sliding_window_sizes:
            attention_config = AttentionConfig(
                qkv_projection_config=linear_config,
                out_projection_config=linear_config,
                query_norm_config=rms_norm_config,
                key_norm_config=rms_norm_config,
                logit_soft_cap=self.attn_logit_softcapping,
                has_sinks=False,
                has_qkv_biases=self.attention_bias,
                has_out_biases=self.attention_bias,
                num_heads=self.num_attention_heads,
                num_groups=self.num_key_value_heads,
                head_dim=self.head_dim,
                is_causal=True,
                scale=attention_scale,
                sliding_window_size=sliding_window_size,
            )
            transformer_layer_config = TransformerLayerConfig(
                pre_mixer_norm_config=rms_norm_config,
                mixer_config=attention_config,
                post_mixer_norm_config=rms_norm_config,
                pre_mlp_norm_config=rms_norm_config,
                mlp_config=mlp_config,
                post_mlp_norm_config=rms_norm_config,
            )
            layer_configs.append(transformer_layer_config)

        transformer_config = TransformerConfig(
            global_rope_config=global_rope_config,
            local_rope_config=local_rope_config,
            layer_configs=tuple(layer_configs),
            output_norm_config=rms_norm_config,
            model_dim=self.hidden_size,
            hidden_dim=self.intermediate_size,
            context_length=context_length or self.max_position_embeddings,
        )

        return DecoderConfig(
            embedding_config=embedding_config,
            transformer_config=transformer_config,
            vocab_size=self.vocab_size,
        )


@dataclass(frozen=True)
class HFGemma3TextConfig(HFGemma3TextConfigRaw, HuggingFaceLMConfig):
    torch_dtype: Literal["bfloat16", "float16", "float32"] = "bfloat16"
    eos_token_id: int | list[int] = field(default_factory=list)


@dataclass(frozen=True)
class HFGemma3VisionConfig:
    hidden_size: int
    image_size: int
    intermediate_size: int
    model_type: Literal["siglip_vision_model"]
    num_attention_heads: int
    num_hidden_layers: int
    patch_size: int
    vision_use_head: bool


@dataclass(frozen=True)
class HFGemma3Config(HuggingFaceLMConfig):
    torch_dtype: Literal["bfloat16", "float16", "float32"]
    architectures: list[Literal["Gemma3ForConditionalGeneration"]]
    boi_token_index: int
    eoi_token_index: int
    eos_token_id: int | list[int]
    image_token_index: int
    initializer_range: float
    mm_tokens_per_image: int
    model_type: Literal["gemma3"]
    text_config: HFGemma3TextConfigRaw
    transformers_version: str
    vision_config: HFGemma3VisionConfig

    def to_decoder_config(
        self,
        context_length: int | None,
        activation_precision: DTypeLike,
        accumulation_precision: DTypeLike,
        metadata_dict: Mapping[str, str],
    ) -> DecoderConfig:
        return self.text_config.to_decoder_config(
            context_length=context_length,
            activation_precision=activation_precision,
            accumulation_precision=accumulation_precision,
            metadata_dict=metadata_dict,
        )



import math
from abc import ABC, abstractmethod
from collections.abc import Mapping, Sequence
from dataclasses import dataclass, replace
from typing import Self

import equinox as eqx
import jax
import jax.numpy as jnp
from einops import rearrange
from jaxtyping import Array, DTypeLike, Float, Int, PRNGKeyArray

from lalamo.common import ParameterTree, dummy_array
from lalamo.quantization import QuantizationMode, dynamically_quantize_activations, quantize_weights
from lalamo.utils import jax_uint4_to_packed_uint8, jax_uint8_to_unpacked_uint4

from .common import (
    LalamoModule,
    register_config_union,
)

__all__ = [
    "FullPrecisionLinear",
    "FullPrecisionLinearConfig",
    "GroupQuantizedLinear",
    "GroupQuantizedLinearConfig",
    "LinearBase",
    "LinearConfig",
    "QLoRALinear",
    "QLoRALinearConfig",
]


class LinearBase[ConfigT: LinearConfigBase](LalamoModule[ConfigT]):
    output_dims: tuple[int, ...] = eqx.field(static=True)

    @property
    @abstractmethod
    def mixture_size(self) -> int | None: ...

    @property
    @abstractmethod
    def input_dim(self) -> int: ...

    @property
    def num_outputs(self) -> int:
        return len(self.output_dims)

    @property
    @abstractmethod
    def has_biases(self) -> bool: ...

    @abstractmethod
    def __call__(
        self,
        inputs: Float[Array, " in_channels"],
    ) -> tuple[Float[Array, " out_channels"], ...]: ...

    def __post_init__(self) -> None:
        assert isinstance(self.output_dims, tuple)

    @staticmethod
    def get_split_points(output_dims: Sequence[int]) -> tuple[int, ...]:
        result = []
        last_split_point = 0
        for dim in output_dims[:-1]:
            last_split_point += dim
            result.append(last_split_point)
        return tuple(result)


@dataclass(frozen=True)
class LinearConfigBase(ABC):
    @abstractmethod
    def random_init(
        self,
        input_dim: int,
        output_dims: tuple[int, ...],
        has_biases: bool,
        *,
        key: PRNGKeyArray,
    ) -> LinearBase: ...

    @abstractmethod
    def random_init_mixture(
        self,
        mixture_size: int,
        input_dim: int,
        output_dims: tuple[int, ...],
        has_biases: bool,
        *,
        key: PRNGKeyArray,
    ) -> LinearBase: ...

    @abstractmethod
    def empty(
        self,
        input_dim: int,
        output_dims: tuple[int, ...],
        has_biases: bool,
    ) -> LinearBase: ...

    @abstractmethod
    def empty_mixture(
        self,
        mixture_size: int,
        input_dim: int,
        output_dims: tuple[int, ...],
        has_biases: bool,
    ) -> LinearBase: ...


@dataclass(frozen=True)
class FullPrecisionLinearConfig(LinearConfigBase):
    precision: DTypeLike

    def random_init(
        self,
        input_dim: int,
        output_dims: tuple[int, ...],
        has_biases: bool,
        *,
        key: PRNGKeyArray,
    ) -> "FullPrecisionLinear":
        scale = 1 / math.sqrt(input_dim)
        weights = jax.random.uniform(
            key,
            (sum(output_dims), input_dim),
            minval=-scale,
            maxval=scale,
            dtype=self.precision,
        )
        if has_biases:
            biases = jnp.zeros((sum(output_dims),), dtype=self.precision)
        else:
            biases = None

        return FullPrecisionLinear(
            config=self,
            output_dims=output_dims,
            weights=weights,
            biases=biases,
        )

    def random_init_mixture(
        self,
        mixture_size: int,
        input_dim: int,
        output_dims: tuple[int, ...],
        has_biases: bool,
        *,
        key: PRNGKeyArray,
    ) -> LinearBase:
        subkeys = jax.random.split(key, mixture_size)
        return eqx.filter_vmap(lambda key: self.random_init(input_dim, output_dims, has_biases, key=key))(subkeys)

    def _empty_general(
        self,
        leading_dims: tuple[int, ...],
        input_dim: int,
        output_dims: tuple[int, ...],
        has_biases: bool,
    ) -> "FullPrecisionLinear":
        weights = dummy_array(
            (*leading_dims, sum(output_dims), input_dim),
            dtype=self.precision,
        )
        if has_biases:
            biases = dummy_array((*leading_dims, sum(output_dims)), dtype=self.precision)
        else:
            biases = None

        return FullPrecisionLinear(
            config=self,
            output_dims=output_dims,
            weights=weights,
            biases=biases,
        )

    def empty(
        self,
        input_dim: int,
        output_dims: tuple[int, ...],
        has_biases: bool,
    ) -> "FullPrecisionLinear":
        return self._empty_general((), input_dim, output_dims, has_biases)

    def empty_mixture(
        self,
        mixture_size: int,
        input_dim: int,
        output_dims: tuple[int, ...],
        has_biases: bool,
    ) -> "FullPrecisionLinear":
        return self._empty_general((mixture_size,), input_dim, output_dims, has_biases)


class FullPrecisionLinear(LinearBase[FullPrecisionLinearConfig]):
    weights: Float[Array, "*components total_out_channels in_channels"]
    biases: Float[Array, "*components total_out_channels"] | None

    @property
    def mixture_size(self) -> int | None:
        match self.weights.shape:
            case [num_components, _, _]:
                return num_components
            case _:
                return None

    @property
    def activation_precision(self) -> DTypeLike:
        return self.config.precision

    @property
    def input_dim(self) -> int:
        *_, _, input_dim = self.weights.shape
        return input_dim

    @property
    def has_biases(self) -> bool:
        return self.biases is not None

    def __post_init__(self) -> None:
        if self.weights.dtype != self.config.precision:
            raise ValueError(
                f"Weight dtype ({self.weights.dtype}) is not equal to specified precision ({self.config.precision}).",
            )
        *w_num_components, w_output_dim, _ = self.weights.shape
        if w_output_dim != sum(self.output_dims):
            raise ValueError(
                f"Number of output channels in weights ({w_output_dim}) is not"
                f" equal to sum of output dims ({sum(self.output_dims)}).",
            )
        if self.biases is None:
            return
        *b_num_components, b_output_dim = self.biases.shape
        if w_output_dim != b_output_dim:
            raise ValueError(
                f"Number of output channels in weights ({w_output_dim}) is not"
                f" equal to number of output channels in biases ({b_output_dim}).",
            )
        if self.biases.dtype != self.config.precision:
            raise ValueError(
                f"Bias dtype ({self.biases.dtype}) is not equal to specified precision ({self.config.precision}).",
            )
        if b_num_components != w_num_components:
            raise ValueError(
                f"Number of mixture components in weights ({w_num_components}) is not"
                f" equal to number of mixture components in biases ({b_num_components}).",
            )

    @eqx.filter_jit
    def __call__(self, inputs: Float[Array, " in_channels"]) -> tuple[Float[Array, " out_channels"], ...]:
        if self.mixture_size is not None:
            raise ValueError(
                "Mixtures of linear layers cannot be called directly."
                "They are intended to be used with methods eqx.filter_vmap or lax.scan instead.",
            )
        result = self.weights @ inputs
        if self.biases is not None:
            result = result + self.biases
        return tuple(jnp.split(result, self.get_split_points(self.output_dims)))

    def export_weights(self) -> ParameterTree:
        result = dict(weights=self.weights)
        if self.biases is not None:
            result["biases"] = self.biases
        return result

    def import_weights(
        self,
        weights: ParameterTree[Array],
    ) -> Self:
        assert isinstance(weights, Mapping)
        return replace(
            self,
            weights=weights["weights"],
            biases=weights["biases"] if self.has_biases else None,
        )


@dataclass(frozen=True)
class QuantizedLinearConfigBase(LinearConfigBase):
    group_size: int
    weight_quantization_mode: QuantizationMode
    activation_quantization_mode: QuantizationMode | None
    activation_precision: DTypeLike


class QuantizedLinearBase[ConfigT: QuantizedLinearConfigBase](LinearBase[ConfigT]):
    biases: Float[Array, "*components total_out_channels"] | None

    @abstractmethod
    def _prepare_scaled_weights(self) -> Float[Array, "*components in_channels total_out_channels"]: ...

    def _apply_weights(self, inputs: Float[Array, " in_channels"]) -> Float[Array, " total_out_channels"]:
        if self.config.activation_quantization_mode is not None:
            inputs = dynamically_quantize_activations(inputs, self.config.activation_quantization_mode)
        return self._prepare_scaled_weights() @ inputs

    @eqx.filter_jit
    def __call__(self, inputs: Float[Array, " in_channels"]) -> tuple[Float[Array, " out_channels"], ...]:
        if self.mixture_size is not None:
            raise ValueError(
                "Mixtures of linear layers cannot be called directly."
                "They are intended to be used with methods eqx.filter_vmap or lax.scan instead.",
            )
        result = self._apply_weights(inputs)
        if self.biases is not None:
            result = result + self.biases
        return tuple(jnp.split(result, self.get_split_points(self.output_dims)))


@dataclass(frozen=True)
class GroupQuantizedLinearConfig(QuantizedLinearConfigBase):
    def random_init(
        self,
        input_dim: int,
        output_dims: tuple[int, ...],
        has_biases: bool,
        *,
        key: PRNGKeyArray,
    ) -> LinearBase:
        min_val, max_val = self.weight_quantization_mode.range
        weights = jax.random.uniform(
            key,
            (sum(output_dims), input_dim),
            minval=min_val - 1,
            maxval=max_val + 1,
            dtype=self.activation_precision,
        )
        num_groups = input_dim // self.group_size
        scale = 1 / ((max_val - min_val) / 2 * math.sqrt(input_dim))
        scales = scale * jnp.ones((sum(output_dims), num_groups), dtype=self.activation_precision)

        if has_biases:
            biases = jnp.zeros((sum(output_dims),), dtype=self.activation_precision)
        else:
            biases = None

        zero_point = min_val + 2 ** (self.weight_quantization_mode.bits - 1)
        zero_points = zero_point * jnp.ones((sum(output_dims), num_groups), dtype=self.activation_precision)

        return GroupQuantizedLinear(
            config=self,
            output_dims=output_dims,
            weights=weights,
            scales=scales,
            zero_points=zero_points,
            biases=biases,
        )

    def random_init_mixture(
        self,
        mixture_size: int,
        input_dim: int,
        output_dims: tuple[int, ...],
        has_biases: bool,
        *,
        key: PRNGKeyArray,
    ) -> LinearBase:
        subkeys = jax.random.split(key, mixture_size)
        return eqx.filter_vmap(lambda key: self.random_init(input_dim, output_dims, has_biases, key=key))(subkeys)

    def _empty_general(
        self,
        leading_dims: tuple[int, ...],
        input_dim: int,
        output_dims: tuple[int, ...],
        has_biases: bool,
    ) -> LinearBase:
        weights = dummy_array(
            (*leading_dims, sum(output_dims), input_dim),
            dtype=self.activation_precision,
        )
        num_groups = input_dim // self.group_size
        scales = dummy_array((*leading_dims, sum(output_dims), num_groups), dtype=self.activation_precision)

        if has_biases:
            biases = dummy_array((*leading_dims, sum(output_dims)), dtype=self.activation_precision)
        else:
            biases = None
        zero_points = dummy_array((*leading_dims, sum(output_dims), num_groups), dtype=self.activation_precision)

        return GroupQuantizedLinear(
            config=self,
            output_dims=output_dims,
            weights=weights,
            scales=scales,
            zero_points=zero_points,
            biases=biases,
        )

    def empty(
        self,
        input_dim: int,
        output_dims: tuple[int, ...],
        has_biases: bool,
    ) -> LinearBase:
        return self._empty_general((), input_dim, output_dims, has_biases)

    def empty_mixture(
        self,
        mixture_size: int,
        input_dim: int,
        output_dims: tuple[int, ...],
        has_biases: bool,
    ) -> LinearBase:
        return self._empty_general((mixture_size,), input_dim, output_dims, has_biases)


class GroupQuantizedLinearBase[ConfigT: GroupQuantizedLinearConfig](QuantizedLinearBase[ConfigT]):
    weights: Float[Array, "*components total_out_channels in_channels"]
    scales: Float[Array, "*components total_out_channels groups"]
    zero_points: Float[Array, "*components total_out_channels groups"]
    biases: Float[Array, "*components total_out_channels"] | None

    @property
    def mixture_size(self) -> int | None:
        match self.weights.shape:
            case [num_components, _, _]:
                return num_components
            case _:
                return None

    @property
    def activation_precision(self) -> DTypeLike:
        return self.config.activation_precision

    @property
    def input_dim(self) -> int:
        *_, _, input_dim = self.weights.shape
        return input_dim

    @property
    def has_biases(self) -> bool:
        return self.biases is not None

    @property
    def num_groups(self) -> int:
        return self.input_dim // self.config.group_size

    @property
    def int_weights(self) -> Int[Array, "*components in_channels out_channels"]:
        quantized = quantize_weights(self.weights, self.config.weight_quantization_mode)
        casted = quantized.astype(self.config.weight_quantization_mode.dtype)

        if self.config.weight_quantization_mode == QuantizationMode.UINT4:
            packed = jax_uint4_to_packed_uint8(casted)
        else:
            packed = casted

        return packed

    @property
    def int_zero_points(self) -> Int[Array, "*components groups out_channels"]:
        quantized = quantize_weights(self.zero_points, self.config.weight_quantization_mode)
        casted = quantized.astype(self.config.weight_quantization_mode.dtype)

        if self.config.weight_quantization_mode == QuantizationMode.UINT4:
            packed = jax_uint4_to_packed_uint8(casted)
        else:
            packed = casted

        return packed

    def __post_init__(self) -> None:  # noqa: PLR0912
        if self.weights.dtype != self.config.activation_precision:
            raise ValueError(
                f"Weight dtype ({self.weights.dtype}) is not equal to specified activation precision"
                f" ({self.config.activation_precision}).",
                " Quantized layers require parameter dtypes to be equal to the activation precision.",
            )
        *w_num_components, w_output_dim, _ = self.weights.shape
        if w_output_dim != sum(self.output_dims):
            raise ValueError(
                f"Number of output channels in weights ({w_output_dim}) is not"
                f" equal to sum of output dims ({sum(self.output_dims)}).",
            )

        if self.scales.dtype != self.config.activation_precision:
            raise ValueError(
                f"Scale dtype ({self.scales.dtype}) is not equal to specified activation precision"
                f" ({self.config.activation_precision}).",
                " Quantized layers require parameter dtypes to be equal to the activation precision.",
            )
        *s_num_components, s_output_dim, s_num_groups = self.scales.shape
        if w_output_dim != s_output_dim:
            raise ValueError(
                f"Number of output channels in weights ({w_output_dim}) is not"
                f" equal to number of output channels in scales ({s_output_dim}).",
            )
        if tuple(s_num_components) != tuple(w_num_components):
            raise ValueError(
                f"Number of mixture components in weights ({w_num_components}) is not"
                f" equal to number of mixture components in scales ({s_num_components}).",
            )
        if s_num_groups != self.num_groups:
            raise ValueError(
                f"Number of groups in scales ({s_num_groups}) is incompatible with"
                f" the specified group size ({self.config.group_size}).",
            )

        if self.zero_points.dtype != self.config.activation_precision:
            raise ValueError(
                f"Zero point dtype ({self.zero_points.dtype}) is not equal to specified activation precision"
                f" ({self.config.activation_precision}).",
                " Quantized layers require parameter dtypes to be equal to the activation precision.",
            )
        *zp_num_components, zp_output_dim, zp_num_groups = self.zero_points.shape
        if w_output_dim != zp_output_dim:
            raise ValueError(
                f"Number of output channels in weights ({w_output_dim}) is not"
                f" equal to number of output channels in zero points ({zp_output_dim}).",
            )
        if tuple(zp_num_components) != tuple(w_num_components):
            raise ValueError(
                f"Number of mixture components in weights ({w_num_components}) is not"
                f" equal to number of mixture components in zero points ({zp_num_components}).",
            )
        if self.num_groups != zp_num_groups:
            raise ValueError(
                f"Number of groups in zero points ({zp_num_groups}) is incompatible with"
                f" the specified group size ({self.config.group_size}).",
            )

        if self.biases is not None:
            if self.biases.dtype != self.config.activation_precision:
                raise ValueError(
                    f"Bias dtype ({self.biases.dtype}) is not equal to specified activation precision"
                    f" ({self.config.activation_precision}).",
                    " Quantized layers require parameter dtypes to be equal to the activation precision.",
                )
            *b_num_components, b_output_dim = self.biases.shape
            if w_output_dim != b_output_dim:
                raise ValueError(
                    f"Number of output channels in weights ({w_output_dim}) is not"
                    f" equal to number of output channels in biases ({b_output_dim}).",
                )
            if tuple(b_num_components) != tuple(w_num_components):
                raise ValueError(
                    f"Number of mixture components in weights ({w_num_components}) is not"
                    f" equal to number of mixture components in biases ({b_num_components}).",
                )

    def _prepare_scaled_weights(self) -> Float[Array, "*components in_channels total_out_channels"]:
        quantized_weights = quantize_weights(self.weights, self.config.weight_quantization_mode)
        grouped_weights = rearrange(
            quantized_weights,
            "... total_out_channels (groups group_channels) -> ... total_out_channels groups group_channels",
            groups=self.num_groups,
        )

        zero_points = rearrange(self.zero_points, "... total_out_channels groups -> ... total_out_channels groups 1")
        grouped_weights = grouped_weights - zero_points

        scales = rearrange(self.scales, "... total_out_channels groups -> ... total_out_channels groups 1")
        scaled_grouped_weights = grouped_weights * scales
        result = rearrange(
            scaled_grouped_weights,
            "... total_out_channels groups group_channels -> ... total_out_channels (groups group_channels)",
        )
        return result

    def export_weights(self) -> ParameterTree:
        result = dict(
            weights=self.int_weights,
            zero_points=self.int_zero_points,
            scales=self.scales,
        )
        if self.biases is not None:
            result["biases"] = self.biases
        return result

    def import_weights(
        self,
        weights: ParameterTree[Array],
    ) -> Self:
        assert isinstance(weights, Mapping)
        assert isinstance(weights["weights"], Array)
        assert isinstance(weights["zero_points"], Array)
        unpacked_weights = weights["weights"]
        unpacked_zero_points = weights["zero_points"]

        if self.config.weight_quantization_mode == QuantizationMode.UINT4:
            unpacked_weights = jax_uint8_to_unpacked_uint4(weights["weights"])
            unpacked_zero_points = jax_uint8_to_unpacked_uint4(weights["zero_points"])

        return replace(
            self,
            weights=unpacked_weights.astype(self.weights.dtype),
            scales=weights["scales"],
            zero_points=unpacked_zero_points.astype(self.zero_points.dtype),
            biases=weights["biases"] if self.has_biases else None,
        )


class GroupQuantizedLinear(GroupQuantizedLinearBase[GroupQuantizedLinearConfig]):
    pass


@dataclass(frozen=True)
class MLXQuantizedLinearConfig(QuantizedLinearConfigBase):
    def random_init(
        self,
        input_dim: int,
        output_dims: tuple[int, ...],
        has_biases: bool,
        *,
        key: PRNGKeyArray,
    ) -> LinearBase:
        min_val, max_val = self.weight_quantization_mode.range
        weights = jax.random.uniform(
            key,
            (sum(output_dims), input_dim),
            minval=min_val - 1,
            maxval=max_val + 1,
            dtype=self.activation_precision,
        )
        num_groups = input_dim // self.group_size
        scale = 1 / ((max_val - min_val) / 2 * math.sqrt(input_dim))
        scales = scale * jnp.ones((sum(output_dims), num_groups), dtype=self.activation_precision)

        if has_biases:
            biases = jnp.zeros((sum(output_dims),), dtype=self.activation_precision)
        else:
            biases = None

        deq_bias = min_val + 2 ** (self.weight_quantization_mode.bits - 1)
        deq_biases = deq_bias * jnp.ones((sum(output_dims), num_groups), dtype=self.activation_precision)

        return MLXQuantizedLinear(
            config=self,
            output_dims=output_dims,
            weights=weights,
            scales=scales,
            deq_biases=deq_biases,
            biases=biases,
        )

    def random_init_mixture(
        self,
        mixture_size: int,
        input_dim: int,
        output_dims: tuple[int, ...],
        has_biases: bool,
        *,
        key: PRNGKeyArray,
    ) -> LinearBase:
        subkeys = jax.random.split(key, mixture_size)
        return eqx.filter_vmap(lambda key: self.random_init(input_dim, output_dims, has_biases, key=key))(subkeys)

    def _empty_general(
        self,
        leading_dims: tuple[int, ...],
        input_dim: int,
        output_dims: tuple[int, ...],
        has_biases: bool,
    ) -> LinearBase:
        weights = dummy_array(
            (*leading_dims, sum(output_dims), input_dim),
            dtype=self.activation_precision,
        )
        num_groups = input_dim // self.group_size
        scales = dummy_array((*leading_dims, sum(output_dims), num_groups), dtype=self.activation_precision)

        if has_biases:
            biases = dummy_array((*leading_dims, sum(output_dims)), dtype=self.activation_precision)
        else:
            biases = None
        deq_biases = dummy_array((*leading_dims, sum(output_dims), num_groups), dtype=self.activation_precision)

        return MLXQuantizedLinear(
            config=self,
            output_dims=output_dims,
            weights=weights,
            scales=scales,
            deq_biases=deq_biases,
            biases=biases,
        )

    def empty(
        self,
        input_dim: int,
        output_dims: tuple[int, ...],
        has_biases: bool,
    ) -> LinearBase:
        return self._empty_general((), input_dim, output_dims, has_biases)

    def empty_mixture(
        self,
        mixture_size: int,
        input_dim: int,
        output_dims: tuple[int, ...],
        has_biases: bool,
    ) -> LinearBase:
        return self._empty_general((mixture_size,), input_dim, output_dims, has_biases)


class MLXQuantizedLinearBase[ConfigT: MLXQuantizedLinearConfig](QuantizedLinearBase[ConfigT]):
    weights: Float[Array, "*components total_out_channels in_channels"]
    scales: Float[Array, "*components total_out_channels groups"]
    deq_biases: Float[Array, "*components total_out_channels groups"]
    biases: Float[Array, "*components total_out_channels"] | None

    @property
    def mixture_size(self) -> int | None:
        match self.weights.shape:
            case [num_components, _, _]:
                return num_components
            case _:
                return None

    @property
    def activation_precision(self) -> DTypeLike:
        return self.config.activation_precision

    @property
    def input_dim(self) -> int:
        *_, _, input_dim = self.weights.shape
        return input_dim

    @property
    def has_biases(self) -> bool:
        return self.biases is not None

    @property
    def num_groups(self) -> int:
        return self.input_dim // self.config.group_size

    @property
    def int_weights(self) -> Int[Array, "*components in_channels out_channels"]:
        quantized = quantize_weights(self.weights, self.config.weight_quantization_mode)
        casted = quantized.astype(self.config.weight_quantization_mode.dtype)

        if self.config.weight_quantization_mode == QuantizationMode.UINT4:
            packed = jax_uint4_to_packed_uint8(casted)
        else:
            packed = casted

        return packed

    def __post_init__(self) -> None:  # noqa: PLR0912
        if self.weights.dtype != self.config.activation_precision:
            raise ValueError(
                f"Weight dtype ({self.weights.dtype}) is not equal to specified activation precision"
                f" ({self.config.activation_precision}).",
                " Quantized layers require parameter dtypes to be equal to the activation precision.",
            )
        *w_num_components, w_output_dim, _ = self.weights.shape
        if w_output_dim != sum(self.output_dims):
            raise ValueError(
                f"Number of output channels in weights ({w_output_dim}) is not"
                f" equal to sum of output dims ({sum(self.output_dims)}).",
            )

        if self.scales.dtype != self.config.activation_precision:
            raise ValueError(
                f"Scale dtype ({self.scales.dtype}) is not equal to specified activation precision"
                f" ({self.config.activation_precision}).",
                " Quantized layers require parameter dtypes to be equal to the activation precision.",
            )
        *s_num_components, s_output_dim, s_num_groups = self.scales.shape
        if w_output_dim != s_output_dim:
            raise ValueError(
                f"Number of output channels in weights ({w_output_dim}) is not"
                f" equal to number of output channels in scales ({s_output_dim}).",
            )
        if tuple(s_num_components) != tuple(w_num_components):
            raise ValueError(
                f"Number of mixture components in weights ({w_num_components}) is not"
                f" equal to number of mixture components in scales ({s_num_components}).",
            )
        if s_num_groups != self.num_groups:
            raise ValueError(
                f"Number of groups in scales ({s_num_groups}) is incompatible with"
                f" the specified group size ({self.config.group_size}).",
            )

        if self.deq_biases.dtype != self.config.activation_precision:
            raise ValueError(
                f"Dequantization bias dtype ({self.deq_biases.dtype}) is not equal to specified activation precision"
                f" ({self.config.activation_precision}).",
                " Quantized layers require parameter dtypes to be equal to the activation precision.",
            )
        *zp_num_components, zp_output_dim, zp_num_groups = self.deq_biases.shape
        if w_output_dim != zp_output_dim:
            raise ValueError(
                f"Number of output channels in weights ({w_output_dim}) is not"
                f" equal to number of output channels in zero points ({zp_output_dim}).",
            )
        if tuple(zp_num_components) != tuple(w_num_components):
            raise ValueError(
                f"Number of mixture components in weights ({w_num_components}) is not"
                f" equal to number of mixture components in zero points ({zp_num_components}).",
            )
        if self.num_groups != zp_num_groups:
            raise ValueError(
                f"Number of groups in zero points ({zp_num_groups}) is incompatible with"
                f" the specified group size ({self.config.group_size}).",
            )

        if self.biases is not None:
            if self.biases.dtype != self.config.activation_precision:
                raise ValueError(
                    f"Bias dtype ({self.biases.dtype}) is not equal to specified activation precision"
                    f" ({self.config.activation_precision}).",
                    " Quantized layers require parameter dtypes to be equal to the activation precision.",
                )
            *b_num_components, b_output_dim = self.biases.shape
            if w_output_dim != b_output_dim:
                raise ValueError(
                    f"Number of output channels in weights ({w_output_dim}) is not"
                    f" equal to number of output channels in biases ({b_output_dim}).",
                )
            if tuple(b_num_components) != tuple(w_num_components):
                raise ValueError(
                    f"Number of mixture components in weights ({w_num_components}) is not"
                    f" equal to number of mixture components in biases ({b_num_components}).",
                )

    def _prepare_scaled_weights(self) -> Float[Array, "*components in_channels total_out_channels"]:
        quantized_weights = quantize_weights(self.weights, self.config.weight_quantization_mode)
        grouped_weights = rearrange(
            quantized_weights,
            "... total_out_channels (groups group_channels) -> ... total_out_channels groups group_channels",
            groups=self.num_groups,
        )

        scales = rearrange(self.scales, "... total_out_channels groups -> ... total_out_channels groups 1")
        deq_biases = rearrange(self.deq_biases, "... total_out_channels groups -> ... total_out_channels groups 1")

        scaled_grouped_weights = grouped_weights * scales + deq_biases
        result = rearrange(
            scaled_grouped_weights,
            "... total_out_channels groups group_channels -> ... total_out_channels (groups group_channels)",
        )
        return result

    def export_weights(self) -> ParameterTree:
        result = dict(
            weights=self.int_weights,
            scales=self.scales,
            deq_biases=self.deq_biases,
        )
        if self.biases is not None:
            result["biases"] = self.biases
        return result

    def import_weights(
        self,
        weights: ParameterTree[Array],
    ) -> Self:
        assert isinstance(weights, Mapping)
        assert isinstance(weights["weights"], Array)

        unpacked_weights = weights["weights"]

        if self.config.weight_quantization_mode == QuantizationMode.UINT4:
            unpacked_weights = jax_uint8_to_unpacked_uint4(weights["weights"])

        return replace(
            self,
            weights=unpacked_weights.astype(self.weights.dtype),
            scales=weights["scales"],
            deq_biases=weights["deq_biases"],
            biases=weights["biases"] if self.has_biases else None,
        )


class MLXQuantizedLinear(MLXQuantizedLinearBase[MLXQuantizedLinearConfig]):
    pass


@dataclass(frozen=True)
class QLoRALinearConfig(GroupQuantizedLinearConfig):
    lora_rank: int
    lora_scale: float
    activation_precision: DTypeLike

    def random_init(
        self,
        input_dim: int,
        output_dims: tuple[int, ...],
        has_biases: bool,
        *,
        key: PRNGKeyArray,
    ) -> LinearBase:
        base_key, derived_key = jax.random.split(key)
        group_quantized_linear = super().random_init(input_dim, output_dims, has_biases, key=base_key)
        assert isinstance(group_quantized_linear, GroupQuantizedLinear)

        down_key, up_key_root = jax.random.split(derived_key)
        hidden_lora_rank = len(output_dims) * self.lora_rank
        max_down_abs_value = 1 / math.sqrt(input_dim)
        lora_down_weights = jax.random.uniform(
            down_key,
            (input_dim, hidden_lora_rank),
            minval=-max_down_abs_value,
            maxval=max_down_abs_value,
            dtype=self.activation_precision,
        )

        up_keys = jax.random.split(up_key_root, len(output_dims))
        max_up_abs_value = 1 / math.sqrt(hidden_lora_rank)
        lora_up_weights = tuple(
            jax.random.uniform(
                up_key,
                (self.lora_rank, output_dim),
                minval=-max_up_abs_value,
                maxval=max_up_abs_value,
                dtype=self.activation_precision,
            )
            for up_key, output_dim in zip(up_keys, output_dims, strict=True)
        )

        return QLoRALinear(
            config=self,
            output_dims=output_dims,
            weights=group_quantized_linear.weights,
            scales=group_quantized_linear.scales,
            biases=group_quantized_linear.biases,
            zero_points=group_quantized_linear.zero_points,
            lora_down_weights=lora_down_weights,
            lora_up_weights=lora_up_weights,
        )

    def random_init_mixture(
        self,
        mixture_size: int,
        input_dim: int,
        output_dims: tuple[int, ...],
        has_biases: bool,
        *,
        key: PRNGKeyArray,
    ) -> LinearBase:
        subkeys = jax.random.split(key, mixture_size)
        return eqx.filter_vmap(lambda k: self.random_init(input_dim, output_dims, has_biases, key=k))(subkeys)

    def empty(
        self,
        input_dim: int,
        output_dims: tuple[int, ...],
        has_biases: bool,
    ) -> LinearBase:
        group_quantized_linear = super().empty(input_dim, output_dims, has_biases)
        assert isinstance(group_quantized_linear, GroupQuantizedLinear)
        hidden_lora_rank = len(output_dims) * self.lora_rank
        lora_down_weights = dummy_array(
            (input_dim, hidden_lora_rank),
            dtype=self.activation_precision,
        )
        lora_up_weights = tuple(
            dummy_array(
                (self.lora_rank, output_dim),
                dtype=self.activation_precision,
            )
            for output_dim in output_dims
        )

        return QLoRALinear(
            config=self,
            output_dims=output_dims,
            weights=group_quantized_linear.weights,
            scales=group_quantized_linear.scales,
            biases=group_quantized_linear.biases,
            zero_points=group_quantized_linear.zero_points,
            lora_down_weights=lora_down_weights,
            lora_up_weights=lora_up_weights,
        )

    def _empty_general(
        self,
        leading_dims: tuple[int, ...],
        input_dim: int,
        output_dims: tuple[int, ...],
        has_biases: bool,
    ) -> LinearBase:
        group_quantized_linear = super().empty(input_dim, output_dims, has_biases)
        assert isinstance(group_quantized_linear, GroupQuantizedLinear)

        hidden_lora_rank = len(output_dims) * self.lora_rank
        lora_down_weights = dummy_array(
            (*leading_dims, input_dim, hidden_lora_rank),
            dtype=self.activation_precision,
        )
        lora_up_weights = tuple(
            dummy_array(
                (*leading_dims, self.lora_rank, output_dim),
                dtype=self.activation_precision,
            )
            for output_dim in output_dims
        )

        return QLoRALinear(
            config=self,
            output_dims=output_dims,
            weights=group_quantized_linear.weights,
            scales=group_quantized_linear.scales,
            biases=group_quantized_linear.biases,
            zero_points=group_quantized_linear.zero_points,
            lora_down_weights=lora_down_weights,
            lora_up_weights=lora_up_weights,
        )

    def empty_mixture(
        self,
        mixture_size: int,
        input_dim: int,
        output_dims: tuple[int, ...],
        has_biases: bool,
    ) -> LinearBase:
        return self._empty_general((mixture_size,), input_dim, output_dims, has_biases)


class QLoRALinear(GroupQuantizedLinearBase[QLoRALinearConfig]):
    lora_down_weights: Float[Array, "*components in_channels total_lora_channels"]
    lora_up_weights: tuple[Float[Array, "*components lora_channels out_channels"], ...]

    def _split_biases(self) -> tuple[Float[Array, "*components out_channels"] | None, ...]:
        if self.biases is not None:
            return tuple(jnp.split(self.biases, self.get_split_points(self.output_dims)))
        return (None,) * len(self.output_dims)

    def __post_init__(self) -> None:
        super().__post_init__()
        if self.lora_down_weights.dtype != self.config.activation_precision:
            raise ValueError(
                f"LORA down weight dtype ({self.lora_down_weights.dtype}) is not equal to the"
                f" specified activation precision ({self.config.activation_precision}).",
                " Quantized layers require parameter dtypes to be equal to the activation precision.",
            )
        *ld_num_components, lora_down_input_dim, lora_down_output_dim = self.lora_down_weights.shape
        if lora_down_output_dim != self.config.lora_rank * self.num_outputs:
            raise ValueError(
                f"Number of output channels in LORA down weights ({lora_down_output_dim}) is not"
                f" equal to lora_rank * num_outputs ({self.config.lora_rank * self.num_outputs}).",
            )
        if lora_down_input_dim != self.input_dim:
            raise ValueError(
                f"Number of input channels in LORA down weights ({lora_down_input_dim}) is not"
                f" equal to input_dim ({self.input_dim}).",
            )
        *w_num_components, _, _ = self.weights.shape
        if tuple(ld_num_components) != tuple(w_num_components):
            raise ValueError(
                f"Number of mixture components in LORA down weights ({ld_num_components}) is not"
                f" equal to number of mixture components in base weights ({w_num_components}).",
            )
        if len(self.lora_up_weights) != self.num_outputs:
            raise ValueError(
                f"Expected {self.num_outputs} LORA up weights, got {len(self.lora_up_weights)}.",
            )
        for lora_up_weight, output_dim in zip(self.lora_up_weights, self.output_dims, strict=True):
            if lora_up_weight.dtype != self.config.activation_precision:
                raise ValueError(
                    f"LORA up weight dtype ({lora_up_weight.dtype}) is not equal to specified activation precision"
                    f" ({self.config.activation_precision}).",
                    " Quantized layers require parameter dtypes to be equal to the activation precision.",
                )
            *lu_num_components, lora_up_input_dim, lora_up_output_dim = lora_up_weight.shape
            if lora_up_output_dim != output_dim:
                raise ValueError(
                    f"Number of output channels in LORA up weights ({lora_up_output_dim}) is not"
                    f" equal to number of output dims ({self.output_dims}).",
                )
            if lora_up_input_dim != self.config.lora_rank:
                raise ValueError(
                    f"Number of input channels in LORA up weights ({lora_up_input_dim}) is not"
                    f" equal to lora_rank ({self.config.lora_rank}).",
                )
            if tuple(lu_num_components) != tuple(w_num_components):
                raise ValueError(
                    f"Number of mixture components in LORA up weights ({lu_num_components}) is not"
                    f" equal to number of mixture components in base weights ({w_num_components}).",
                )

    @eqx.filter_jit
    def __call__(self, inputs: Float[Array, " in_channels"]) -> tuple[Float[Array, " out_channels"], ...]:
        if self.mixture_size is not None:
            raise ValueError(
                "Mixtures of linear layers cannot be called directly."
                "They are intended to be used with methods eqx.filter_vmap or lax.scan instead.",
            )
        joint_q_out = self._apply_weights(inputs)
        q_outs = jnp.split(joint_q_out, self.get_split_points(self.output_dims))

        joint_lora_hidden = inputs @ self.lora_down_weights
        lora_hiddens = jnp.split(joint_lora_hidden, self.get_split_points([self.config.lora_rank] * self.num_outputs))
        lora_outs = [
            lora_hidden @ lora_up_weight
            for lora_up_weight, lora_hidden in zip(self.lora_up_weights, lora_hiddens, strict=True)
        ]

        results = []
        for q_out, lora_out, bias in zip(q_outs, lora_outs, self._split_biases(), strict=True):
            result = q_out + self.config.lora_scale * lora_out
            if bias is not None:
                result = result + bias
            results.append(result)

        return tuple(results)

    def export_weights(self) -> ParameterTree:
        quantized_linear_weights: dict[str, ParameterTree] = super().export_weights()  # type: ignore
        return dict(
            down_weights=self.lora_down_weights,
            up_weights=self.lora_up_weights,
            **quantized_linear_weights,
        )

    def import_weights(
        self,
        weights: ParameterTree[Array],
    ) -> Self:
        base = super().import_weights(weights)
        assert isinstance(weights, Mapping)
        assert isinstance(weights["up_weights"], Sequence)
        return replace(
            base,
            lora_down_weights=weights["down_weights"],
            lora_up_weights=tuple(up_weights for up_weights in weights["up_weights"]),
        )


LinearConfig = FullPrecisionLinearConfig | GroupQuantizedLinearConfig | MLXQuantizedLinearConfig | QLoRALinearConfig


register_config_union(LinearConfig)  # type: ignore (pyright bug)
